@@ -2,6 +2,6 @@
 async fn main() -> Result<(), deltalake::DeltaTableError> {
     let table_path = "./tests/data/delta-0.8.0";
     let table = deltalake::open_table(table_path).await?;
-    println!("{}", table);
+    println!("{table}");
     Ok(())
 }
