@@ -6,8 +6,8 @@ use serde_json::json;
 use std::collections::HashMap;
 
 use crate::action::{
-    Action, ActionError, Add, ColumnCountStat, ColumnValueStat, MetaData, Protocol, Remove, Stats,
-    Txn,
+    Action, ActionError, Add, AddCDCFile, ColumnCountStat, ColumnValueStat, MetaData, Protocol,
+    Remove, Stats, Txn,
 };
 
 fn populate_hashmap_with_option_from_parquet_map(
@@ -30,11 +30,19 @@ fn populate_hashmap_with_option_from_parquet_map(
 
 fn gen_action_type_error(action: &str, field: &str, expected_type: &str) -> ActionError {
     ActionError::InvalidField(format!(
-        "type for {} in {} action should be {}",
-        field, action, expected_type
+        "type for {field} in {action} action should be {expected_type}"
     ))
 }
 
+impl AddCDCFile {
+    fn from_parquet_record(_record: &parquet::record::Row) -> Result<Self, ActionError> {
+        let re = Self {
+            ..Default::default()
+        };
+        Ok(re)
+    }
+}
+
 impl Add {
     fn from_parquet_record(record: &parquet::record::Row) -> Result<Self, ActionError> {
         let mut re = Self {
@@ -74,8 +82,7 @@ impl Add {
                     )
                     .map_err(|estr| {
                         ActionError::InvalidField(format!(
-                            "Invalid partitionValues for add action: {}",
-                            estr,
+                            "Invalid partitionValues for add action: {estr}",
                         ))
                     })?;
                 }
@@ -95,8 +102,7 @@ impl Add {
                         populate_hashmap_with_option_from_parquet_map(&mut tags, tags_map)
                             .map_err(|estr| {
                                 ActionError::InvalidField(format!(
-                                    "Invalid tags for add action: {}",
-                                    estr,
+                                    "Invalid tags for add action: {estr}",
                                 ))
                             })?;
                         re.tags = Some(tags);
@@ -122,7 +128,7 @@ impl Add {
                     }
                 },
                 _ => {
-                    log::warn!(
+                    log::debug!(
                         "Unexpected field name `{}` for add action: {:?}",
                         name,
                         record
@@ -142,53 +148,46 @@ impl Add {
 
                 for (i, (name, _)) in record.get_column_iter().enumerate() {
                     match name.as_str() {
-                        "numRecords" => match record.get_long(i) {
-                            Ok(v) => {
+                        "numRecords" => if let Ok(v) = record.get_long(i) {
                                 stats.num_records = v;
-                            }
-                            _ => {
+                            } else {
                                 log::error!("Expect type of stats_parsed field numRecords to be long, got: {}", record);
                             }
-                        }
-                        "minValues" => match record.get_group(i) {
-                            Ok(row) => {
-                                for (name, field) in row.get_column_iter() {
-                                    stats.min_values.insert(name.clone(), field.into());
+                        "minValues" => if let Ok(row) = record.get_group(i) {
+                            for (name, field) in row.get_column_iter() {
+                                if !matches!(field, Field::Null) {
+                                    if let Some(values) = field_to_value_stat(field, name) {
+                                        stats.min_values.insert(name.clone(), values);
+                                    }
                                 }
                             }
-                            _ => {
-                                log::error!("Expect type of stats_parsed field minRecords to be struct, got: {}", record);
-                            }
+                        } else {
+                            log::error!("Expect type of stats_parsed field minRecords to be struct, got: {}", record);
                         }
-                        "maxValues" => match record.get_group(i) {
-                            Ok(row) => {
-                                for (name, field) in row.get_column_iter() {
-                                    stats.max_values.insert(name.clone(), field.into());
+                        "maxValues" => if let Ok(row) = record.get_group(i) {
+                            for (name, field) in row.get_column_iter() {
+                                if !matches!(field, Field::Null) {
+                                    if let Some(values) = field_to_value_stat(field, name) {
+                                        stats.max_values.insert(name.clone(), values);
+                                    }
                                 }
                             }
-                            _ => {
-                                log::error!("Expect type of stats_parsed field maxRecords to be struct, got: {}", record);
-                            }
+                        } else {
+                            log::error!("Expect type of stats_parsed field maxRecords to be struct, got: {}", record);
                         }
-                        "nullCount" => match record.get_group(i) {
-                            Ok(row) => {
-                                for (name, field) in row.get_column_iter() {
-                                    match field.try_into() {
-                                        Ok(count) => {
-                                            stats.null_count.insert(name.clone(), count);
-                                        },
-                                        _ => {
-                                            log::warn!("Expect type of nullCount field to be struct or int64, got: {}", field);
-                                        },
-                                    };
+                        "nullCount" => if let Ok(row) = record.get_group(i) {
+                            for (name, field) in row.get_column_iter() {
+                                if !matches!(field, Field::Null) {
+                                    if let Some(count) = field_to_count_stat(field, name) {
+                                        stats.null_count.insert(name.clone(), count);
+                                    }
                                 }
                             }
-                            _ => {
-                                log::error!("Expect type of stats_parsed field nullCount to be struct, got: {}", record);
-                            }
+                        } else {
+                            log::error!("Expect type of stats_parsed field nullCount to be struct, got: {}", record);
                         }
                         _ => {
-                            log::warn!(
+                            log::debug!(
                                 "Unexpected field name `{}` for stats_parsed: {:?}",
                                 name,
                                 record,
@@ -202,80 +201,89 @@ impl Add {
     }
 }
 
-impl From<&Field> for ColumnValueStat {
-    fn from(field: &Field) -> Self {
-        match field {
-            Field::Group(group) => ColumnValueStat::Column(HashMap::from_iter(
-                group
-                    .get_column_iter()
-                    .map(|(field_name, field)| (field_name.clone(), field.into())),
-            )),
-            _ => ColumnValueStat::Value(primitive_parquet_field_to_json_value(field)),
-        }
-    }
-}
-
-impl TryFrom<&Field> for ColumnCountStat {
-    type Error = &'static str;
-
-    fn try_from(field: &Field) -> Result<Self, Self::Error> {
-        match field {
-            Field::Group(group) => Ok(ColumnCountStat::Column(HashMap::from_iter(
-                group
-                    .get_column_iter()
-                    .filter_map(|(field_name, field)| match field.try_into() {
-                        Ok(value) => Some((field_name.clone(), value)),
-                        _ => {
-                            log::warn!(
-                                "Unexpected type when parsing nullCounts for {}. Found {}",
-                                field_name,
-                                field
-                            );
-                            None
-                        }
-                    }),
-            ))),
-            Field::Long(value) => Ok(ColumnCountStat::Value(*value)),
-            _ => Err("Invalid type for nullCounts"),
-        }
-    }
-}
-
-fn primitive_parquet_field_to_json_value(field: &Field) -> serde_json::Value {
+fn field_to_value_stat(field: &Field, field_name: &str) -> Option<ColumnValueStat> {
     match field {
-        Field::Null => serde_json::Value::Null,
-        Field::Bool(value) => json!(value),
-        Field::Byte(value) => json!(value),
-        Field::Short(value) => json!(value),
-        Field::Int(value) => json!(value),
-        Field::Long(value) => json!(value),
-        Field::Float(value) => json!(value),
-        Field::Double(value) => json!(value),
-        Field::Str(value) => json!(value),
-        Field::Decimal(decimal) => match BigInt::from_signed_bytes_be(decimal.data()).to_f64() {
-            Some(int) => json!(int / (10_i64.pow((decimal.scale()).try_into().unwrap()) as f64)),
-            _ => serde_json::Value::Null,
-        },
-        Field::TimestampMillis(timestamp) => {
-            serde_json::Value::String(convert_timestamp_millis_to_string(*timestamp))
+        Field::Group(group) => {
+            let values = group.get_column_iter().filter_map(|(name, sub_field)| {
+                field_to_value_stat(sub_field, name).map(|val| (name.clone(), val))
+            });
+            Some(ColumnValueStat::Column(HashMap::from_iter(values)))
         }
-        Field::Date(date) => serde_json::Value::String(convert_date_to_string(*date)),
         _ => {
-            log::warn!("Unexpected field type {:?}", field,);
-            serde_json::Value::Null
+            if let Ok(val) = primitive_parquet_field_to_json_value(field) {
+                Some(ColumnValueStat::Value(val))
+            } else {
+                log::warn!(
+                    "Unexpected type when parsing min/max values for {}. Found {}",
+                    field_name,
+                    field
+                );
+                None
+            }
         }
     }
 }
 
-fn convert_timestamp_millis_to_string(value: u64) -> String {
-    let dt = Utc.timestamp((value / 1000) as i64, ((value % 1000) * 1000000) as u32);
-    dt.to_rfc3339_opts(SecondsFormat::Millis, true)
+fn field_to_count_stat(field: &Field, field_name: &str) -> Option<ColumnCountStat> {
+    match field {
+        Field::Group(group) => {
+            let counts = group.get_column_iter().filter_map(|(name, sub_field)| {
+                field_to_count_stat(sub_field, name).map(|count| (name.clone(), count))
+            });
+            Some(ColumnCountStat::Column(HashMap::from_iter(counts)))
+        }
+        Field::Long(value) => Some(ColumnCountStat::Value(*value)),
+        _ => {
+            log::warn!(
+                "Unexpected type when parsing nullCounts for {}. Found {}",
+                field_name,
+                field
+            );
+            None
+        }
+    }
 }
 
-fn convert_date_to_string(value: u32) -> String {
+fn primitive_parquet_field_to_json_value(field: &Field) -> Result<serde_json::Value, &'static str> {
+    match field {
+        Field::Bool(value) => Ok(json!(value)),
+        Field::Byte(value) => Ok(json!(value)),
+        Field::Short(value) => Ok(json!(value)),
+        Field::Int(value) => Ok(json!(value)),
+        Field::Long(value) => Ok(json!(value)),
+        Field::Float(value) => Ok(json!(value)),
+        Field::Double(value) => Ok(json!(value)),
+        Field::Str(value) => Ok(json!(value)),
+        Field::Decimal(decimal) => match BigInt::from_signed_bytes_be(decimal.data()).to_f64() {
+            Some(int) => Ok(json!(
+                int / (10_i64.pow((decimal.scale()).try_into().unwrap()) as f64)
+            )),
+            _ => Err("Invalid type for min/max values."),
+        },
+        Field::TimestampMillis(timestamp) => Ok(serde_json::Value::String(
+            convert_timestamp_millis_to_string(*timestamp)?,
+        )),
+        Field::Date(date) => Ok(serde_json::Value::String(convert_date_to_string(*date)?)),
+        _ => Err("Invalid type for min/max values."),
+    }
+}
+
+fn convert_timestamp_millis_to_string(value: i64) -> Result<String, &'static str> {
+    let dt = Utc
+        .timestamp_opt(value / 1000, ((value % 1000) * 1000000) as u32)
+        .single()
+        .ok_or("Value out of bounds")?;
+    Ok(dt.to_rfc3339_opts(SecondsFormat::Millis, true))
+}
+
+fn convert_date_to_string(value: i32) -> Result<String, &'static str> {
     static NUM_SECONDS_IN_DAY: i64 = 60 * 60 * 24;
-    let dt = Utc.timestamp(value as i64 * NUM_SECONDS_IN_DAY, 0).date();
-    format!("{}", dt.format("%Y-%m-%d"))
+    let dt = Utc
+        .timestamp_opt(value as i64 * NUM_SECONDS_IN_DAY, 0)
+        .single()
+        .ok_or("Value out of bounds")?
+        .date_naive();
+    Ok(format!("{}", dt.format("%Y-%m-%d")))
 }
 
 impl MetaData {
@@ -341,8 +349,7 @@ impl MetaData {
                     )
                     .map_err(|estr| {
                         ActionError::InvalidField(format!(
-                            "Invalid configuration for metaData action: {}",
-                            estr,
+                            "Invalid configuration for metaData action: {estr}",
                         ))
                     })?;
                 }
@@ -366,8 +373,7 @@ impl MetaData {
                             )
                             .map_err(|estr| {
                                 ActionError::InvalidField(format!(
-                                    "Invalid format.options for metaData action: {}",
-                                    estr,
+                                    "Invalid format.options for metaData action: {estr}",
                                 ))
                             })?;
                             re.format.options = options;
@@ -378,7 +384,7 @@ impl MetaData {
                     }
                 }
                 _ => {
-                    log::warn!(
+                    log::debug!(
                         "Unexpected field name `{}` for metaData action: {:?}",
                         name,
                         record
@@ -432,8 +438,7 @@ impl Remove {
                         )
                         .map_err(|estr| {
                             ActionError::InvalidField(format!(
-                                "Invalid partitionValues for remove action: {}",
-                                estr,
+                                "Invalid partitionValues for remove action: {estr}",
                             ))
                         })?;
                         re.partition_values = Some(partition_values);
@@ -446,8 +451,7 @@ impl Remove {
                         populate_hashmap_with_option_from_parquet_map(&mut tags, tags_map)
                             .map_err(|estr| {
                                 ActionError::InvalidField(format!(
-                                    "Invalid tags for remove action: {}",
-                                    estr,
+                                    "Invalid tags for remove action: {estr}",
                                 ))
                             })?;
                         re.tags = Some(tags);
@@ -461,7 +465,7 @@ impl Remove {
                 }
                 "numRecords" => {}
                 _ => {
-                    log::warn!(
+                    log::debug!(
                         "Unexpected field name `{}` for remove action: {:?}",
                         name,
                         record
@@ -497,7 +501,7 @@ impl Txn {
                     re.last_updated = record.get_long(i).map(Some).unwrap_or(None);
                 }
                 _ => {
-                    log::warn!(
+                    log::debug!(
                         "Unexpected field name `{}` for txn action: {:?}",
                         name,
                         record
@@ -529,7 +533,7 @@ impl Protocol {
                     })?;
                 }
                 _ => {
-                    log::warn!(
+                    log::debug!(
                         "Unexpected field name `{}` for protocol action: {:?}",
                         name,
                         record
@@ -584,10 +588,10 @@ impl Action {
             "remove" => Action::remove(Remove::from_parquet_record(col_data)?),
             "txn" => Action::txn(Txn::from_parquet_record(col_data)?),
             "protocol" => Action::protocol(Protocol::from_parquet_record(col_data)?),
+            "cdc" => Action::cdc(AddCDCFile::from_parquet_record(col_data)?),
             name => {
                 return Err(ActionError::InvalidField(format!(
-                    "Unexpected action from checkpoint: {}",
-                    name,
+                    "Unexpected action from checkpoint: {name}",
                 )));
             }
         })
