@@ -140,13 +140,13 @@ async fn create_checkpoint_for(
     let size = parquet_bytes.len() as i64;
     let checkpoint = CheckPoint::new(version, size, None);
 
-    let file_name = format!("{:020}.checkpoint.parquet", version);
+    let file_name = format!("{version:020}.checkpoint.parquet");
     let checkpoint_path = storage.log_path().child(file_name);
 
     debug!("Writing checkpoint to {:?}.", checkpoint_path);
     storage.put(&checkpoint_path, parquet_bytes).await?;
 
-    let last_checkpoint_content: Value = serde_json::to_value(&checkpoint)?;
+    let last_checkpoint_content: Value = serde_json::to_value(checkpoint)?;
     let last_checkpoint_content = bytes::Bytes::from(serde_json::to_vec(&last_checkpoint_content)?);
 
     debug!("Writing _last_checkpoint to {:?}.", last_checkpoint_path);
@@ -359,13 +359,13 @@ fn parquet_bytes_from_state(state: &DeltaTableState) -> Result<bytes::Bytes, Che
 
         // As a "new writer", we should always set `extendedFileMetadata` when writing, and include/ignore the other three fields accordingly.
         // https://github.com/delta-io/delta/blob/fb0452c2fb142310211c6d3604eefb767bb4a134/core/src/main/scala/org/apache/spark/sql/delta/actions/actions.scala#L311-L314
-        if None == r.extended_file_metadata {
+        if r.extended_file_metadata.is_none() {
             r.extended_file_metadata = Some(false);
         }
 
         action::Action::remove(r)
     }))
-    .map(|a| serde_json::to_value(a).map_err(ArrowError::from))
+    .map(|a| serde_json::to_value(a).map_err(|err| ArrowError::JsonError(err.to_string())))
     // adds
     .chain(state.files().iter().map(|f| {
         checkpoint_add_from_state(f, partition_col_data_types.as_slice(), &stats_conversions)
@@ -398,7 +398,8 @@ fn checkpoint_add_from_state(
     partition_col_data_types: &[(&str, &SchemaDataType)],
     stats_conversions: &[(SchemaPath, SchemaDataType)],
 ) -> Result<Value, ArrowError> {
-    let mut v = serde_json::to_value(action::Action::add(add.clone()))?;
+    let mut v = serde_json::to_value(action::Action::add(add.clone()))
+        .map_err(|err| ArrowError::JsonError(err.to_string()))?;
 
     v["add"]["dataChange"] = Value::Bool(false);
 
@@ -413,12 +414,14 @@ fn checkpoint_add_from_state(
             }
         }
 
-        let partition_values_parsed = serde_json::to_value(partition_values_parsed)?;
+        let partition_values_parsed = serde_json::to_value(partition_values_parsed)
+            .map_err(|err| ArrowError::JsonError(err.to_string()))?;
         v["add"]["partitionValues_parsed"] = partition_values_parsed;
     }
 
     if let Ok(Some(stats)) = add.get_stats() {
-        let mut stats = serde_json::to_value(stats)?;
+        let mut stats =
+            serde_json::to_value(stats).map_err(|err| ArrowError::JsonError(err.to_string()))?;
         let min_values = stats.get_mut("minValues").and_then(|v| v.as_object_mut());
 
         if let Some(min_values) = min_values {
