@@ -5,18 +5,20 @@
 use std::collections::HashMap;
 use std::convert::TryFrom;
 use std::fmt;
+use std::fmt::Formatter;
 use std::io::{BufRead, BufReader, Cursor};
+use std::sync::Arc;
 use std::{cmp::max, cmp::Ordering, collections::HashSet};
 
 use super::action;
 use super::action::{Action, DeltaOperation};
-use super::partitions::{DeltaTablePartition, PartitionFilter};
+use super::partitions::PartitionFilter;
 use super::schema::*;
 use super::table_state::DeltaTableState;
 use crate::action::{Add, Stats};
 use crate::delta_config::DeltaConfigError;
-use crate::storage::ObjectStoreRef;
-use crate::vacuum::{Vacuum, VacuumError};
+use crate::operations::vacuum::VacuumBuilder;
+use crate::storage::{commit_uri_from_version, ObjectStoreRef};
 
 use chrono::{DateTime, Duration, Utc};
 use futures::StreamExt;
@@ -24,10 +26,13 @@ use lazy_static::lazy_static;
 use log::debug;
 use object_store::{path::Path, Error as ObjectStoreError, ObjectStore};
 use regex::Regex;
-use serde::{Deserialize, Serialize};
+use serde::de::{Error, SeqAccess, Visitor};
+use serde::ser::SerializeSeq;
+use serde::{Deserialize, Deserializer, Serialize, Serializer};
 use serde_json::{Map, Value};
 use uuid::Uuid;
 
+// TODO re-exports only for transition
 pub use crate::builder::{DeltaTableBuilder, DeltaTableConfig, DeltaVersion};
 
 /// Metadata for a checkpoint file
@@ -85,22 +90,21 @@ pub enum DeltaTableError {
         #[from]
         source: ObjectStoreError,
     },
-    /// Error returned when reading the checkpoint failed.
-    #[cfg(feature = "parquet")]
-    #[error("Failed to read checkpoint: {}", .source)]
-    Parquet {
-        /// Parquet error details returned when reading the checkpoint failed.
-        #[from]
-        source: parquet::errors::ParquetError,
-    },
-    /// Error returned when parsing checkpoint parquet using parquet2 crate.
-    #[cfg(feature = "parquet2")]
+
+    /// Error returned when parsing checkpoint parquet.
+    #[cfg(any(feature = "parquet", feature = "parquet2"))]
     #[error("Failed to parse parquet: {}", .source)]
     Parquet {
-        /// Parquet error details returned when parsing the checkpoint parquet
+        /// Parquet error details returned when reading the checkpoint failed.
+        #[cfg(feature = "parquet")]
+        #[from]
+        source: parquet::errors::ParquetError,
+        /// Parquet error details returned when reading the checkpoint failed.
+        #[cfg(feature = "parquet2")]
         #[from]
         source: parquet2::error::Error,
     },
+
     /// Error returned when converting the schema in Arrow format failed.
     #[cfg(feature = "arrow")]
     #[error("Failed to convert into Arrow schema: {}", .source)]
@@ -109,12 +113,30 @@ pub enum DeltaTableError {
         #[from]
         source: arrow::error::ArrowError,
     },
+
     /// Error returned when the log record has an invalid JSON.
-    #[error("Invalid JSON in log record: {}", .source)]
-    InvalidJson {
-        /// JSON error details returned when the log record has an invalid JSON.
-        #[from]
-        source: serde_json::error::Error,
+    #[error("Invalid JSON in log record, version={}, line=`{}`, err=`{}`", .version, .line, .json_err)]
+    InvalidJsonLog {
+        /// JSON error details returned when parsing the record JSON.
+        json_err: serde_json::error::Error,
+        /// invalid log entry content.
+        line: String,
+        /// corresponding table version for the log file.
+        version: DeltaDataTypeVersion,
+    },
+    /// Error returned when the log contains invalid stats JSON.
+    #[error("Invalid JSON in file stats: {}", .json_err)]
+    InvalidStatsJson {
+        /// JSON error details returned when parsing the stats JSON.
+        json_err: serde_json::error::Error,
+    },
+    /// Error returned when the log contains invalid stats JSON.
+    #[error("Invalid JSON in invariant expression, line=`{line}`, err=`{json_err}`")]
+    InvalidInvariantJson {
+        /// JSON error details returned when parsing the invariant expression JSON.
+        json_err: serde_json::error::Error,
+        /// Invariant expression.
+        line: String,
     },
     /// Error returned when the DeltaTable has an invalid version.
     #[error("Invalid table version: {0}")]
@@ -194,9 +216,38 @@ pub enum DeltaTableError {
     /// Error returned when user attempts to commit actions that don't belong to the next version.
     #[error("Delta transaction failed, version {0} does not follow {1}")]
     VersionMismatch(DeltaDataTypeVersion, DeltaDataTypeVersion),
+    /// A Feature is missing to perform operation
+    #[error("Delta-rs must be build with feature '{feature}' to support loading from: {url}.")]
+    MissingFeature {
+        /// Name of the missiing feature
+        feature: &'static str,
+        /// Storage location url
+        url: String,
+    },
+    /// A Feature is missing to perform operation
+    #[error("Cannot infer storage location from: {0}")]
+    InvalidTableLocation(String),
+    /// Generic Delta Table error
+    #[error("Log JSON serialization error: {json_err}")]
+    SerializeLogJson {
+        /// JSON serialization error
+        json_err: serde_json::error::Error,
+    },
+    /// Generic Delta Table error
+    #[error("Schema JSON serialization error: {json_err}")]
+    SerializeSchemaJson {
+        /// JSON serialization error
+        json_err: serde_json::error::Error,
+    },
     /// Generic Delta Table error
     #[error("Generic DeltaTable error: {0}")]
     Generic(String),
+    /// Generic Delta Table error
+    #[error("Generic error: {source}")]
+    GenericError {
+        /// Source error
+        source: Box<dyn std::error::Error + Send + Sync + 'static>,
+    },
 }
 
 /// Delta table metadata
@@ -299,24 +350,6 @@ impl TryFrom<action::MetaData> for DeltaTableMetaData {
     }
 }
 
-impl TryFrom<DeltaTableMetaData> for action::MetaData {
-    type Error = serde_json::error::Error;
-
-    fn try_from(metadata: DeltaTableMetaData) -> Result<Self, Self::Error> {
-        let schema_string = serde_json::to_string(&metadata.schema)?;
-        Ok(Self {
-            id: metadata.id,
-            name: metadata.name,
-            description: metadata.description,
-            format: metadata.format,
-            schema_string,
-            partition_columns: metadata.partition_columns,
-            created_time: metadata.created_time,
-            configuration: metadata.configuration,
-        })
-    }
-}
-
 /// Error related to Delta log application
 #[derive(thiserror::Error, Debug)]
 pub enum ApplyLogError {
@@ -324,7 +357,7 @@ pub enum ApplyLogError {
     #[error("End of transaction log")]
     EndOfLog,
     /// Error returned when the JSON of the log record is invalid.
-    #[error("Invalid JSON in log record")]
+    #[error("Invalid JSON found when applying log record")]
     InvalidJson {
         /// JSON error details returned when reading the JSON log record.
         #[from]
@@ -415,12 +448,78 @@ pub struct DeltaTable {
     pub state: DeltaTableState,
     /// the load options used during load
     pub config: DeltaTableConfig,
+    /// object store to access log and data files
     pub(crate) storage: ObjectStoreRef,
-
+    /// file metadata for latest checkpoint
     last_check_point: Option<CheckPoint>,
+    /// table versions associated with timestamps
     version_timestamp: HashMap<DeltaDataTypeVersion, i64>,
 }
 
+impl Serialize for DeltaTable {
+    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        let mut seq = serializer.serialize_seq(None)?;
+        seq.serialize_element(&self.state)?;
+        seq.serialize_element(&self.config)?;
+        seq.serialize_element(self.storage.as_ref())?;
+        seq.serialize_element(&self.last_check_point)?;
+        seq.serialize_element(&self.version_timestamp)?;
+        seq.end()
+    }
+}
+
+impl<'de> Deserialize<'de> for DeltaTable {
+    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
+    where
+        D: Deserializer<'de>,
+    {
+        struct DeltaTableVisitor {}
+
+        impl<'de> Visitor<'de> for DeltaTableVisitor {
+            type Value = DeltaTable;
+
+            fn expecting(&self, formatter: &mut Formatter) -> fmt::Result {
+                formatter.write_str("struct DeltaTable")
+            }
+
+            fn visit_seq<A>(self, mut seq: A) -> Result<Self::Value, A::Error>
+            where
+                A: SeqAccess<'de>,
+            {
+                let state = seq
+                    .next_element()?
+                    .ok_or_else(|| A::Error::invalid_length(0, &self))?;
+                let config = seq
+                    .next_element()?
+                    .ok_or_else(|| A::Error::invalid_length(0, &self))?;
+                let storage = seq
+                    .next_element()?
+                    .ok_or_else(|| A::Error::invalid_length(0, &self))?;
+                let last_check_point = seq
+                    .next_element()?
+                    .ok_or_else(|| A::Error::invalid_length(0, &self))?;
+                let version_timestamp = seq
+                    .next_element()?
+                    .ok_or_else(|| A::Error::invalid_length(0, &self))?;
+
+                let table = DeltaTable {
+                    state,
+                    config,
+                    storage: Arc::new(storage),
+                    last_check_point,
+                    version_timestamp,
+                };
+                Ok(table)
+            }
+        }
+
+        deserializer.deserialize_seq(DeltaTableVisitor {})
+    }
+}
+
 impl DeltaTable {
     /// Create a new Delta Table struct without loading any data from backing storage.
     ///
@@ -436,22 +535,31 @@ impl DeltaTable {
         }
     }
 
+    /// Create a new [`DeltaTable`] from a [`DeltaTableState`] without loading any
+    /// data from backing storage.
+    ///
+    /// NOTE: This is for advanced users. If you don't know why you need to use this method,
+    /// please call one of the `open_table` helper methods instead.
+    pub(crate) fn new_with_state(storage: ObjectStoreRef, state: DeltaTableState) -> Self {
+        Self {
+            state,
+            storage,
+            config: Default::default(),
+            last_check_point: None,
+            version_timestamp: HashMap::new(),
+        }
+    }
+
     /// get a shared reference to the delta object store
     pub fn object_store(&self) -> ObjectStoreRef {
         self.storage.clone()
     }
 
-    /// The
+    /// The URI of the underlying data
     pub fn table_uri(&self) -> String {
         self.storage.root_uri()
     }
 
-    /// Return the uri of commit version.
-    pub fn commit_uri_from_version(&self, version: DeltaDataTypeVersion) -> Path {
-        let version = format!("{:020}.json", version);
-        Path::from_iter(["_delta_log", &version])
-    }
-
     /// Return the list of paths of given checkpoint.
     pub fn get_checkpoint_data_paths(&self, check_point: &CheckPoint) -> Vec<Path> {
         let checkpoint_prefix = format!("{:020}", check_point.version);
@@ -460,7 +568,7 @@ impl DeltaTable {
 
         match check_point.parts {
             None => {
-                let path = log_path.child(&*format!("{}.checkpoint.parquet", checkpoint_prefix));
+                let path = log_path.child(&*format!("{checkpoint_prefix}.checkpoint.parquet"));
                 checkpoint_data_paths.push(path);
             }
             Some(parts) => {
@@ -509,6 +617,7 @@ impl DeltaTable {
 
     async fn get_last_checkpoint(&self) -> Result<CheckPoint, LoadCheckpointError> {
         let last_checkpoint_path = Path::from_iter(["_delta_log", "_last_checkpoint"]);
+        debug!("loading checkpoint from {last_checkpoint_path}");
         match self.storage.get(&last_checkpoint_path).await {
             Ok(data) => Ok(serde_json::from_slice(&data.bytes().await?)?),
             Err(ObjectStoreError::NotFound { .. }) => {
@@ -589,17 +698,7 @@ impl DeltaTable {
         Ok(cp)
     }
 
-    async fn apply_log(&mut self, version: DeltaDataTypeVersion) -> Result<(), ApplyLogError> {
-        let new_state = DeltaTableState::from_commit(self, version).await?;
-        self.state.merge(
-            new_state,
-            self.config.require_tombstones,
-            self.config.require_files,
-        );
-
-        Ok(())
-    }
-
+    #[cfg(any(feature = "parquet", feature = "parquet2"))]
     async fn restore_checkpoint(&mut self, check_point: CheckPoint) -> Result<(), DeltaTableError> {
         self.state = DeltaTableState::from_checkpoint(self, &check_point).await?;
 
@@ -618,13 +717,11 @@ impl DeltaTable {
             }
         };
 
+        debug!("start with latest checkpoint version: {version}");
+
         // scan logs after checkpoint
         loop {
-            match self
-                .storage
-                .head(&self.commit_uri_from_version(version))
-                .await
-            {
+            match self.storage.head(&commit_uri_from_version(version)).await {
                 Ok(meta) => {
                     // also cache timestamp for version
                     self.version_timestamp
@@ -653,7 +750,7 @@ impl DeltaTable {
         Ok(version)
     }
 
-    /// Currently loaded evrsion of the table
+    /// Currently loaded version of the table
     pub fn version(&self) -> DeltaDataTypeVersion {
         self.state.version()
     }
@@ -671,7 +768,7 @@ impl DeltaTable {
         current_version: DeltaDataTypeVersion,
     ) -> Result<PeekCommit, DeltaTableError> {
         let next_version = current_version + 1;
-        let commit_uri = self.commit_uri_from_version(next_version);
+        let commit_uri = commit_uri_from_version(next_version);
         let commit_log_bytes = self.storage.get(&commit_uri).await;
         let commit_log_bytes = match commit_log_bytes {
             Err(ObjectStoreError::NotFound { .. }) => return Ok(PeekCommit::UpToDate),
@@ -679,59 +776,74 @@ impl DeltaTable {
             Ok(result) => result.bytes().await,
         }?;
 
+        debug!("parsing commit with version {next_version}...");
         let reader = BufReader::new(Cursor::new(commit_log_bytes));
 
         let mut actions = Vec::new();
-        for line in reader.lines() {
-            let action: action::Action = serde_json::from_str(line?.as_str())?;
+        for re_line in reader.lines() {
+            let line = re_line?;
+            let lstr = line.as_str();
+            let action =
+                serde_json::from_str(lstr).map_err(|e| DeltaTableError::InvalidJsonLog {
+                    json_err: e,
+                    version: next_version,
+                    line,
+                })?;
             actions.push(action);
         }
         Ok(PeekCommit::New(next_version, actions))
     }
 
-    ///Apply any actions associated with the PeekCommit to the DeltaTable
-    pub fn apply_actions(
-        &mut self,
-        new_version: DeltaDataTypeVersion,
-        actions: Vec<Action>,
-    ) -> Result<(), DeltaTableError> {
-        if self.version() + 1 != new_version {
-            return Err(DeltaTableError::VersionMismatch(
-                new_version,
-                self.version(),
-            ));
-        }
-
-        let s = DeltaTableState::from_actions(actions, new_version)?;
-        self.state
-            .merge(s, self.config.require_tombstones, self.config.require_files);
-
-        Ok(())
-    }
-
     /// Updates the DeltaTable to the most recent state committed to the transaction log by
     /// loading the last checkpoint and incrementally applying each version since.
+    #[cfg(any(feature = "parquet", feature = "parquet2"))]
     pub async fn update(&mut self) -> Result<(), DeltaTableError> {
         match self.get_last_checkpoint().await {
             Ok(last_check_point) => {
+                debug!("update with latest checkpoint {last_check_point:?}");
                 if Some(last_check_point) == self.last_check_point {
-                    self.update_incremental().await
+                    self.update_incremental(None).await
                 } else {
                     self.last_check_point = Some(last_check_point);
                     self.restore_checkpoint(last_check_point).await?;
-                    self.update_incremental().await
+                    self.update_incremental(None).await
                 }
             }
-            Err(LoadCheckpointError::NotFound) => self.update_incremental().await,
-            Err(e) => Err(DeltaTableError::LoadCheckpoint { source: e }),
+            Err(LoadCheckpointError::NotFound) => {
+                debug!("update without checkpoint");
+                self.update_incremental(None).await
+            }
+            Err(source) => Err(DeltaTableError::LoadCheckpoint { source }),
         }
     }
 
+    /// Updates the DeltaTable to the most recent state committed to the transaction log.
+    #[cfg(not(any(feature = "parquet", feature = "parquet2")))]
+    pub async fn update(&mut self) -> Result<(), DeltaTableError> {
+        self.update_incremental(None).await
+    }
+
     /// Updates the DeltaTable to the latest version by incrementally applying newer versions.
     /// It assumes that the table is already updated to the current version `self.version`.
-    pub async fn update_incremental(&mut self) -> Result<(), DeltaTableError> {
-        while let PeekCommit::New(version, actions) = self.peek_next_commit(self.version()).await? {
-            self.apply_actions(version, actions)?;
+    pub async fn update_incremental(
+        &mut self,
+        max_version: Option<DeltaDataTypeVersion>,
+    ) -> Result<(), DeltaTableError> {
+        debug!(
+            "incremental update with version({}) and max_version({max_version:?})",
+            self.version(),
+        );
+
+        while let PeekCommit::New(new_version, actions) =
+            self.peek_next_commit(self.version()).await?
+        {
+            debug!("merging table state with version: {new_version}");
+            let s = DeltaTableState::from_actions(actions, new_version)?;
+            self.state
+                .merge(s, self.config.require_tombstones, self.config.require_files);
+            if Some(self.version()) == max_version {
+                return Ok(());
+            }
         }
 
         if self.version() == -1 {
@@ -751,7 +863,7 @@ impl DeltaTable {
         version: DeltaDataTypeVersion,
     ) -> Result<(), DeltaTableError> {
         // check if version is valid
-        let commit_uri = self.commit_uri_from_version(version);
+        let commit_uri = commit_uri_from_version(version);
         match self.storage.head(&commit_uri).await {
             Ok(_) => {}
             Err(ObjectStoreError::NotFound { .. }) => {
@@ -762,25 +874,21 @@ impl DeltaTable {
             }
         }
 
-        let mut next_version;
         // 1. find latest checkpoint below version
+        #[cfg(any(feature = "parquet", feature = "parquet2"))]
         match self.find_latest_check_point_for_version(version).await? {
             Some(check_point) => {
                 self.restore_checkpoint(check_point).await?;
-                next_version = check_point.version + 1;
             }
             None => {
                 // no checkpoint found, clear table state and start from the beginning
-                self.state = DeltaTableState::with_version(0);
-                next_version = 0;
+                self.state = DeltaTableState::with_version(-1);
             }
         }
 
+        debug!("update incrementally from version {version}");
         // 2. apply all logs starting from checkpoint
-        while next_version <= version {
-            self.apply_log(next_version).await?;
-            next_version += 1;
-        }
+        self.update_incremental(Some(version)).await?;
 
         Ok(())
     }
@@ -792,10 +900,7 @@ impl DeltaTable {
         match self.version_timestamp.get(&version) {
             Some(ts) => Ok(*ts),
             None => {
-                let meta = self
-                    .storage
-                    .head(&self.commit_uri_from_version(version))
-                    .await?;
+                let meta = self.storage.head(&commit_uri_from_version(version)).await?;
                 let ts = meta.last_modified.timestamp();
                 // also cache timestamp for version
                 self.version_timestamp.insert(version, ts);
@@ -812,7 +917,7 @@ impl DeltaTable {
     pub async fn history(
         &mut self,
         limit: Option<usize>,
-    ) -> Result<Vec<Map<String, Value>>, DeltaTableError> {
+    ) -> Result<Vec<action::CommitInfo>, DeltaTableError> {
         let mut version = match limit {
             Some(l) => max(self.version() - l as i64 + 1, 0),
             None => self.get_earliest_delta_log_version().await?,
@@ -864,35 +969,7 @@ impl DeltaTable {
         &'a self,
         filters: &'a [PartitionFilter<'a, &'a str>],
     ) -> Result<impl Iterator<Item = &'a Add> + '_, DeltaTableError> {
-        let current_metadata = self
-            .state
-            .current_metadata()
-            .ok_or(DeltaTableError::NoMetadata)?;
-        if !filters
-            .iter()
-            .all(|f| current_metadata.partition_columns.contains(&f.key.into()))
-        {
-            return Err(DeltaTableError::InvalidPartitionFilter {
-                partition_filter: format!("{:?}", filters),
-            });
-        }
-
-        let partition_col_data_types: HashMap<&str, &SchemaDataType> = current_metadata
-            .get_partition_col_data_types()
-            .into_iter()
-            .collect();
-
-        let actions = self.state.files().iter().filter(move |add| {
-            let partitions = add
-                .partition_values
-                .iter()
-                .map(|p| DeltaTablePartition::from_partition_value(p, ""))
-                .collect::<Vec<DeltaTablePartition>>();
-            filters
-                .iter()
-                .all(|filter| filter.match_partitions(&partitions, &partition_col_data_types))
-        });
-        Ok(actions)
+        self.state.get_active_add_actions_by_partitions(filters)
     }
 
     /// Returns the file list tracked in current table state filtered by provided
@@ -922,41 +999,33 @@ impl DeltaTable {
     /// Returns an iterator of file names present in the loaded state
     #[inline]
     pub fn get_files_iter(&self) -> impl Iterator<Item = Path> + '_ {
-        self.state
-            .files()
-            .iter()
-            .map(|add| Path::from(add.path.as_ref()))
+        self.state.file_paths_iter()
     }
 
     /// Returns a collection of file names present in the loaded state
     #[inline]
     pub fn get_files(&self) -> Vec<Path> {
-        self.get_files_iter().collect()
+        self.state.file_paths_iter().collect()
     }
 
     /// Returns file names present in the loaded state in HashSet
     pub fn get_file_set(&self) -> HashSet<Path> {
-        self.state
-            .files()
-            .iter()
-            .map(|add| Path::from(add.path.as_ref()))
-            .collect()
+        self.state.file_paths_iter().collect()
     }
 
     /// Returns a URIs for all active files present in the current table version.
     pub fn get_file_uris(&self) -> impl Iterator<Item = String> + '_ {
         self.state
-            .files()
-            .iter()
-            .map(|add| self.storage.to_uri(&Path::from(add.path.as_ref())))
+            .file_paths_iter()
+            .map(|path| self.storage.to_uri(&path))
     }
 
     /// Returns statistics for files, in order
     pub fn get_stats(&self) -> impl Iterator<Item = Result<Option<Stats>, DeltaTableError>> + '_ {
-        self.state
-            .files()
-            .iter()
-            .map(|add| add.get_stats().map_err(DeltaTableError::from))
+        self.state.files().iter().map(|add| {
+            add.get_stats()
+                .map_err(|e| DeltaTableError::InvalidStatsJson { json_err: e })
+        })
     }
 
     /// Returns partition values for files, in order
@@ -1000,28 +1069,10 @@ impl DeltaTable {
         self.state.min_writer_version()
     }
 
-    /// Vacuum the delta table see [`Vacuum`] for more info
-    pub async fn vacuum(
-        &mut self,
-        retention_hours: Option<u64>,
-        dry_run: bool,
-        enforce_retention_duration: bool,
-    ) -> Result<Vec<String>, VacuumError> {
-        let mut plan = Vacuum::default()
-            .dry_run(dry_run)
-            .enforce_retention_duration(enforce_retention_duration);
-        if let Some(hours) = retention_hours {
-            plan = plan.with_retention_period(Duration::hours(hours as i64));
-        }
-
-        let res = plan.execute(self).await?;
-        Ok(res.files_deleted)
-    }
-
     /// Return table schema parsed from transaction log. Return None if table hasn't been loaded or
     /// no metadata was found in the log.
     pub fn schema(&self) -> Option<&Schema> {
-        self.state.current_metadata().map(|m| &m.schema)
+        self.state.schema()
     }
 
     /// Return table schema parsed from transaction log. Return `DeltaTableError` if table hasn't
@@ -1039,6 +1090,25 @@ impl DeltaTable {
             .get_configuration())
     }
 
+    /// Vacuum the delta table. See [`VacuumBuilder`] for more information.
+    pub async fn vacuum(
+        &mut self,
+        retention_hours: Option<u64>,
+        dry_run: bool,
+        enforce_retention_duration: bool,
+    ) -> Result<Vec<String>, DeltaTableError> {
+        let mut plan = VacuumBuilder::new(self.object_store(), self.state.clone())
+            .with_dry_run(dry_run)
+            .with_enforce_retention_duration(enforce_retention_duration);
+        if let Some(hours) = retention_hours {
+            plan = plan.with_retention_period(Duration::hours(hours as i64));
+        }
+
+        let (table, metrics) = plan.await?;
+        self.state = table.state;
+        Ok(metrics.files_deleted)
+    }
+
     /// Creates a new DeltaTransaction for the DeltaTable.
     /// The transaction holds a mutable reference to the DeltaTable, preventing other references
     /// until the transaction is dropped.
@@ -1062,7 +1132,7 @@ impl DeltaTable {
         // move temporary commit file to delta log directory
         // rely on storage to fail if the file already exists -
         self.storage
-            .rename_if_not_exists(&commit.uri, &self.commit_uri_from_version(version))
+            .rename_if_not_exists(&commit.uri, &commit_uri_from_version(version))
             .await
             .map_err(|e| match e {
                 ObjectStoreError::AlreadyExists { .. } => {
@@ -1098,7 +1168,7 @@ impl DeltaTable {
         );
 
         let mut actions = vec![
-            Action::commitInfo(enriched_commit_info),
+            Action::commit_info(enriched_commit_info),
             Action::protocol(protocol),
             Action::metaData(meta),
         ];
@@ -1171,7 +1241,7 @@ impl fmt::Display for DeltaTable {
         writeln!(f, "\tversion: {}", self.version())?;
         match self.state.current_metadata() {
             Some(metadata) => {
-                writeln!(f, "\tmetadata: {}", metadata)?;
+                writeln!(f, "\tmetadata: {metadata}")?;
             }
             None => {
                 writeln!(f, "\tmetadata: None")?;
@@ -1308,32 +1378,30 @@ impl<'a> DeltaTransaction<'a> {
             .iter()
             .any(|a| matches!(a, action::Action::commitInfo(..)))
         {
-            let mut commit_info = Map::<String, Value>::new();
-            commit_info.insert(
-                "timestamp".to_string(),
-                Value::Number(serde_json::Number::from(Utc::now().timestamp_millis())),
-            );
-            commit_info.insert(
+            let mut extra_info = Map::<String, Value>::new();
+            let mut commit_info = operation.map(|op| op.get_commit_info()).unwrap_or_default();
+            commit_info.timestamp = Some(Utc::now().timestamp_millis());
+            extra_info.insert(
                 "clientVersion".to_string(),
                 Value::String(format!("delta-rs.{}", crate_version())),
             );
-
-            if let Some(op) = &operation {
-                commit_info.append(&mut op.get_commit_info())
-            }
             if let Some(mut meta) = app_metadata {
-                commit_info.append(&mut meta)
+                extra_info.append(&mut meta)
             }
+            commit_info.info = extra_info;
             self.add_action(action::Action::commitInfo(commit_info));
         }
 
         // Serialize all actions that are part of this log entry.
-        let log_entry = bytes::Bytes::from(log_entry_from_actions(&self.actions)?);
+        let log_entry = bytes::Bytes::from(
+            log_entry_from_actions(&self.actions)
+                .map_err(|e| DeltaTableError::SerializeLogJson { json_err: e })?,
+        );
 
         // Write delta log entry as temporary file to storage. For the actual commit,
         // the temporary file is moved (atomic rename) to the delta log folder within `commit` function.
         let token = Uuid::new_v4().to_string();
-        let file_name = format!("_commit_{}.json.tmp", token);
+        let file_name = format!("_commit_{token}.json.tmp");
         let path = Path::from_iter(["_delta_log", &file_name]);
 
         self.delta_table.storage.put(&path, log_entry).await?;
@@ -1406,6 +1474,19 @@ pub async fn open_table(table_uri: impl AsRef<str>) -> Result<DeltaTable, DeltaT
     Ok(table)
 }
 
+/// Same as `open_table`, but also accepts storage options to aid in building the table for a deduced
+/// `StorageService`.
+pub async fn open_table_with_storage_options(
+    table_uri: impl AsRef<str>,
+    storage_options: HashMap<String, String>,
+) -> Result<DeltaTable, DeltaTableError> {
+    let table = DeltaTableBuilder::from_uri(table_uri)
+        .with_storage_options(storage_options)
+        .load()
+        .await?;
+    Ok(table)
+}
+
 /// Creates a DeltaTable from the given path and loads it with the metadata from the given version.
 /// Infers the storage backend to use from the scheme in the given table path.
 pub async fn open_table_with_version(
@@ -1441,17 +1522,29 @@ pub fn crate_version() -> &'static str {
 #[cfg(test)]
 mod tests {
     use super::*;
+    use crate::action::Protocol;
     use crate::builder::DeltaTableBuilder;
     use pretty_assertions::assert_eq;
     use std::io::{BufRead, BufReader};
     use std::{collections::HashMap, fs::File, path::Path};
+    use tempdir::TempDir;
 
-    #[cfg(any(feature = "s3", feature = "s3-rustls"))]
+    #[tokio::test]
+    async fn table_round_trip() {
+        let (_, _, dt, tmp_dir) = create_test_table().await;
+        let bytes = serde_json::to_vec(&dt).unwrap();
+        let actual: DeltaTable = serde_json::from_slice(&bytes).unwrap();
+        assert_eq!(actual.version(), dt.version());
+        drop(tmp_dir);
+    }
+
+    #[cfg(any(feature = "s3", feature = "s3-native-tls"))]
     #[test]
-    fn normalize_table_uri() {
+    fn normalize_table_uri_s3() {
+        std::env::set_var("AWS_DEFAULT_REGION", "us-east-1");
         for table_uri in [
             "s3://tests/data/delta-0.8.0/",
-            // "s3://tests/data/delta-0.8.0//",
+            "s3://tests/data/delta-0.8.0//",
             "s3://tests/data/delta-0.8.0",
         ]
         .iter()
@@ -1461,8 +1554,7 @@ mod tests {
         }
     }
 
-    #[tokio::test]
-    async fn test_create_delta_table() {
+    async fn create_test_table() -> (DeltaTableMetaData, Protocol, DeltaTable, TempDir) {
         // Setup
         let test_schema = Schema::new(vec![
             SchemaField::new(
@@ -1493,7 +1585,7 @@ mod tests {
             min_writer_version: 1,
         };
 
-        let tmp_dir = tempdir::TempDir::new("create_table_test").unwrap();
+        let tmp_dir = TempDir::new("create_table_test").unwrap();
         let table_dir = tmp_dir.path().join("test_create");
         std::fs::create_dir(&table_dir).unwrap();
 
@@ -1514,6 +1606,12 @@ mod tests {
         dt.create(delta_md.clone(), protocol.clone(), Some(commit_info), None)
             .await
             .unwrap();
+        (delta_md, protocol, dt, tmp_dir)
+    }
+
+    #[tokio::test]
+    async fn test_create_delta_table() {
+        let (delta_md, protocol, dt, tmp_dir) = create_test_table().await;
 
         // Validation
         // assert DeltaTable version is now 0 and no data files have been added
@@ -1545,27 +1643,7 @@ mod tests {
                     assert_eq!(DeltaTableMetaData::try_from(action).unwrap(), delta_md);
                 }
                 Action::commitInfo(action) => {
-                    let mut modified_action = action;
-                    let timestamp = serde_json::Number::from(0i64);
-                    modified_action["timestamp"] = Value::Number(serde_json::Number::from(0i64));
-                    let mut expected = Map::<String, Value>::new();
-                    expected.insert(
-                        "operation".to_string(),
-                        serde_json::Value::String("CREATE TABLE".to_string()),
-                    );
-                    expected.insert(
-                        "userName".to_string(),
-                        serde_json::Value::String("test user".to_string()),
-                    );
-                    expected.insert(
-                        "delta-rs".to_string(),
-                        serde_json::Value::String(crate_version().to_string()),
-                    );
-                    expected.insert(
-                        "timestamp".to_string(),
-                        serde_json::Value::Number(timestamp),
-                    );
-                    assert_eq!(modified_action, expected)
+                    assert_eq!(action.operation, Some("CREATE TABLE".to_string()));
                 }
                 _ => (),
             }
@@ -1576,5 +1654,6 @@ mod tests {
         let current_metadata = dt.get_metadata().unwrap();
         assert!(current_metadata.partition_columns.is_empty());
         assert!(current_metadata.configuration.is_empty());
+        drop(tmp_dir);
     }
 }
