@@ -8,8 +8,6 @@ use arrow::datatypes::{
 use arrow::error::ArrowError;
 use lazy_static::lazy_static;
 use regex::Regex;
-use std::collections::BTreeMap;
-use std::collections::HashMap;
 use std::convert::TryFrom;
 
 impl TryFrom<&schema::Schema> for ArrowSchema {
@@ -30,24 +28,20 @@ impl TryFrom<&schema::SchemaField> for ArrowField {
     type Error = ArrowError;
 
     fn try_from(f: &schema::SchemaField) -> Result<Self, ArrowError> {
-        let mut field = ArrowField::new(
+        let metadata = f
+            .get_metadata()
+            .iter()
+            .map(|(key, val)| Ok((key.clone(), serde_json::to_string(val)?)))
+            .collect::<Result<_, serde_json::Error>>()
+            .map_err(|err| ArrowError::JsonError(err.to_string()))?;
+
+        let field = ArrowField::new(
             f.get_name(),
             ArrowDataType::try_from(f.get_type())?,
             f.is_nullable(),
-        );
+        )
+        .with_metadata(metadata);
 
-        let metadata: Option<BTreeMap<String, String>> = Some(f.get_metadata())
-            .filter(|metadata| metadata.is_empty())
-            .map(|metadata| {
-                metadata
-                    .iter()
-                    .map(|(key, val)| Ok((key.clone(), serde_json::to_string(val)?)))
-                    .collect::<Result<_, serde_json::Error>>()
-                    .map_err(|err| ArrowError::JsonError(err.to_string()))
-            })
-            .transpose()?;
-
-        field.set_metadata(metadata);
         Ok(field)
     }
 }
@@ -106,18 +100,16 @@ impl TryFrom<&schema::SchemaDataType> for ArrowDataType {
                     decimal if DECIMAL_REGEX.is_match(decimal) => {
                         let extract = DECIMAL_REGEX.captures(decimal).ok_or_else(|| {
                             ArrowError::SchemaError(format!(
-                                "Invalid decimal type for Arrow: {}",
-                                decimal
+                                "Invalid decimal type for Arrow: {decimal}"
                             ))
                         })?;
                         let precision = extract.get(1).and_then(|v| v.as_str().parse::<u8>().ok());
-                        let scale = extract.get(2).and_then(|v| v.as_str().parse::<u8>().ok());
+                        let scale = extract.get(2).and_then(|v| v.as_str().parse::<i8>().ok());
                         match (precision, scale) {
                             // TODO how do we decide which variant (128 / 256) to use?
                             (Some(p), Some(s)) => Ok(ArrowDataType::Decimal128(p, s)),
                             _ => Err(ArrowError::SchemaError(format!(
-                                "Invalid precision or scale decimal type for Arrow: {}",
-                                decimal
+                                "Invalid precision or scale decimal type for Arrow: {decimal}"
                             ))),
                         }
                     }
@@ -131,8 +123,7 @@ impl TryFrom<&schema::SchemaDataType> for ArrowDataType {
                         Ok(ArrowDataType::Timestamp(TimeUnit::Microsecond, None))
                     }
                     s => Err(ArrowError::SchemaError(format!(
-                        "Invalid data type for Arrow: {}",
-                        s
+                        "Invalid data type for Arrow: {s}"
                     ))),
                 }
             }
@@ -205,12 +196,9 @@ impl TryFrom<&ArrowField> for schema::SchemaField {
             arrow_field.is_nullable(),
             arrow_field
                 .metadata()
-                .as_ref()
-                .map_or_else(HashMap::new, |m| {
-                    m.iter()
-                        .map(|(k, v)| (k.clone(), serde_json::Value::String(v.clone())))
-                        .collect()
-                }),
+                .iter()
+                .map(|(k, v)| (k.clone(), serde_json::Value::String(v.clone())))
+                .collect(),
         ))
     }
 }
@@ -229,17 +217,20 @@ impl TryFrom<&ArrowDataType> for schema::SchemaDataType {
             ArrowDataType::Boolean => Ok(schema::SchemaDataType::primitive("boolean".to_string())),
             ArrowDataType::Binary => Ok(schema::SchemaDataType::primitive("binary".to_string())),
             ArrowDataType::Decimal128(p, s) => Ok(schema::SchemaDataType::primitive(format!(
-                "decimal({},{})",
-                p, s
+                "decimal({p},{s})"
             ))),
             ArrowDataType::Decimal256(p, s) => Ok(schema::SchemaDataType::primitive(format!(
-                "decimal({},{})",
-                p, s
+                "decimal({p},{s})"
             ))),
             ArrowDataType::Date32 => Ok(schema::SchemaDataType::primitive("date".to_string())),
             ArrowDataType::Timestamp(TimeUnit::Microsecond, None) => {
                 Ok(schema::SchemaDataType::primitive("timestamp".to_string()))
             }
+            ArrowDataType::Timestamp(TimeUnit::Microsecond, Some(tz))
+                if tz.eq_ignore_ascii_case("utc") =>
+            {
+                Ok(schema::SchemaDataType::primitive("timestamp".to_string()))
+            }
             ArrowDataType::Struct(fields) => {
                 let converted_fields: Result<Vec<schema::SchemaField>, _> =
                     fields.iter().map(|field| field.try_into()).collect();
@@ -274,8 +265,7 @@ impl TryFrom<&ArrowDataType> for schema::SchemaDataType {
                 }
             }
             s => Err(ArrowError::SchemaError(format!(
-                "Invalid data type for Delta Lake: {}",
-                s
+                "Invalid data type for Delta Lake: {s}"
             ))),
         }
     }
@@ -463,7 +453,7 @@ pub(crate) fn delta_log_schema_for_table(
             .for_each(|f| max_min_schema_for_fields(&mut max_min_vec, f));
 
         stats_parsed_fields.extend(
-            ["minValues", "maxValues"].iter().map(|name| {
+            ["minValues", "maxValues"].into_iter().map(|name| {
                 ArrowField::new(name, ArrowDataType::Struct(max_min_vec.clone()), true)
             }),
         );
@@ -693,7 +683,7 @@ mod tests {
     fn test_arrow_from_delta_decimal_type() {
         let precision = 20;
         let scale = 2;
-        let decimal_type = format!["decimal({p},{s})", p = precision, s = scale];
+        let decimal_type = format!["decimal({precision},{scale})"];
         let decimal_field = crate::SchemaDataType::primitive(decimal_type);
         assert_eq!(
             <ArrowDataType as TryFrom<&crate::SchemaDataType>>::try_from(&decimal_field).unwrap(),
@@ -705,11 +695,8 @@ mod tests {
     fn test_arrow_from_delta_wrong_decimal_type() {
         let precision = 20;
         let scale = "wrong";
-        let decimal_type = format!["decimal({p},{s})", p = precision, s = scale];
-        let _error = format!(
-            "Invalid precision or scale decimal type for Arrow: {}",
-            scale
-        );
+        let decimal_type = format!["decimal({precision},{scale})"];
+        let _error = format!("Invalid precision or scale decimal type for Arrow: {scale}");
         let decimal_field = crate::SchemaDataType::primitive(decimal_type);
         assert!(matches!(
             <ArrowDataType as TryFrom<&crate::SchemaDataType>>::try_from(&decimal_field)
@@ -717,4 +704,132 @@ mod tests {
             arrow::error::ArrowError::SchemaError(_error),
         ));
     }
+
+    #[test]
+    fn test_arrow_from_delta_timestamp_type() {
+        let timestamp_field = crate::SchemaDataType::primitive("timestamp".to_string());
+        assert_eq!(
+            <ArrowDataType as TryFrom<&crate::SchemaDataType>>::try_from(&timestamp_field).unwrap(),
+            ArrowDataType::Timestamp(TimeUnit::Microsecond, None)
+        );
+    }
+
+    #[test]
+    fn test_delta_from_arrow_timestamp_type() {
+        let timestamp_field = ArrowDataType::Timestamp(TimeUnit::Microsecond, None);
+        assert_eq!(
+            <crate::SchemaDataType as TryFrom<&ArrowDataType>>::try_from(&timestamp_field).unwrap(),
+            crate::SchemaDataType::primitive("timestamp".to_string())
+        );
+    }
+
+    #[test]
+    fn test_delta_from_arrow_timestamp_type_with_tz() {
+        let timestamp_field =
+            ArrowDataType::Timestamp(TimeUnit::Microsecond, Some("UTC".to_string()));
+        assert_eq!(
+            <crate::SchemaDataType as TryFrom<&ArrowDataType>>::try_from(&timestamp_field).unwrap(),
+            crate::SchemaDataType::primitive("timestamp".to_string())
+        );
+    }
+
+    #[test]
+    fn test_max_min_schema_for_fields() {
+        let mut max_min_vec: Vec<ArrowField> = Vec::new();
+        let fields = [
+            ArrowField::new("simple", ArrowDataType::Int32, true),
+            ArrowField::new(
+                "struct",
+                ArrowDataType::Struct(vec![ArrowField::new("simple", ArrowDataType::Int32, true)]),
+                true,
+            ),
+            ArrowField::new(
+                "list",
+                ArrowDataType::List(Box::new(ArrowField::new(
+                    "simple",
+                    ArrowDataType::Int32,
+                    true,
+                ))),
+                true,
+            ),
+            ArrowField::new(
+                "map",
+                ArrowDataType::Map(
+                    Box::new(ArrowField::new(
+                        "struct",
+                        ArrowDataType::Struct(vec![
+                            ArrowField::new("key", ArrowDataType::Int32, true),
+                            ArrowField::new("value", ArrowDataType::Int32, true),
+                        ]),
+                        true,
+                    )),
+                    true,
+                ),
+                true,
+            ),
+        ];
+
+        let expected = vec![fields[0].clone(), fields[1].clone()];
+
+        fields
+            .iter()
+            .for_each(|f| max_min_schema_for_fields(&mut max_min_vec, f));
+
+        assert_eq!(max_min_vec, expected);
+    }
+
+    #[test]
+    fn test_null_count_schema_for_fields() {
+        let mut null_count_vec: Vec<ArrowField> = Vec::new();
+        let fields = [
+            ArrowField::new("int32", ArrowDataType::Int32, true),
+            ArrowField::new("int64", ArrowDataType::Int64, true),
+            ArrowField::new("Utf8", ArrowDataType::Utf8, true),
+            ArrowField::new(
+                "list",
+                ArrowDataType::List(Box::new(ArrowField::new(
+                    "simple",
+                    ArrowDataType::Int32,
+                    true,
+                ))),
+                true,
+            ),
+            ArrowField::new(
+                "map",
+                ArrowDataType::Map(
+                    Box::new(ArrowField::new(
+                        "struct",
+                        ArrowDataType::Struct(vec![
+                            ArrowField::new("key", ArrowDataType::Int32, true),
+                            ArrowField::new("value", ArrowDataType::Int32, true),
+                        ]),
+                        true,
+                    )),
+                    true,
+                ),
+                true,
+            ),
+            ArrowField::new(
+                "struct",
+                ArrowDataType::Struct(vec![ArrowField::new("int32", ArrowDataType::Int32, true)]),
+                true,
+            ),
+        ];
+        let expected = vec![
+            ArrowField::new(fields[0].name(), ArrowDataType::Int64, true),
+            ArrowField::new(fields[1].name(), ArrowDataType::Int64, true),
+            ArrowField::new(fields[2].name(), ArrowDataType::Int64, true),
+            ArrowField::new(fields[3].name(), ArrowDataType::Int64, true),
+            ArrowField::new(fields[4].name(), ArrowDataType::Int64, true),
+            ArrowField::new(
+                fields[5].name(),
+                ArrowDataType::Struct(vec![ArrowField::new("int32", ArrowDataType::Int64, true)]),
+                true,
+            ),
+        ];
+        fields
+            .iter()
+            .for_each(|f| null_count_schema_for_fields(&mut null_count_vec, f));
+        assert_eq!(null_count_vec, expected);
+    }
 }
