@@ -1,5 +1,6 @@
 //! AWS S3 storage backend.
 
+use super::utils::str_is_truthy;
 use crate::builder::{s3_storage_options, str_option};
 use bytes::Bytes;
 use dynamodb_lock::{DynamoError, LockClient, LockItem, DEFAULT_MAX_RETRY_ACQUIRE_LOCK_ATTEMPTS};
@@ -80,7 +81,11 @@ enum S3LockError {
     #[error("Rename target already exists")]
     AlreadyExists,
 
-    #[error("Atomic rename requires a LockClient for S3 backends.")]
+    #[error(
+        "Atomic rename requires a LockClient for S3 backends. \
+         Either configure the LockClient, or set AWS_S3_ALLOW_UNSAFE_RENAME=true \
+         to opt out of support for concurrent writers."
+    )]
     LockClientRequired,
 }
 
@@ -98,7 +103,7 @@ impl From<S3LockError> for ObjectStoreError {
 pub struct LockData {
     /// Source object key
     pub source: String,
-    /// Destination object ket
+    /// Destination object key
     pub destination: String,
 }
 
@@ -222,9 +227,11 @@ impl S3LockClient {
 pub struct S3StorageOptions {
     pub endpoint_url: Option<String>,
     pub region: Region,
+    pub profile: Option<String>,
     pub aws_access_key_id: Option<String>,
     pub aws_secret_access_key: Option<String>,
     pub aws_session_token: Option<String>,
+    pub virtual_hosted_style_request: bool,
     pub locking_provider: Option<String>,
     pub assume_role_arn: Option<String>,
     pub assume_role_session_name: Option<String>,
@@ -232,12 +239,13 @@ pub struct S3StorageOptions {
     pub s3_pool_idle_timeout: Duration,
     pub sts_pool_idle_timeout: Duration,
     pub s3_get_internal_server_error_retries: usize,
+    pub allow_unsafe_rename: bool,
     pub extra_opts: HashMap<String, String>,
 }
 
 impl S3StorageOptions {
     /// Creates an instance of S3StorageOptions from the given HashMap.
-    pub fn from_map(options: HashMap<String, String>) -> S3StorageOptions {
+    pub fn from_map(options: &HashMap<String, String>) -> S3StorageOptions {
         let extra_opts = options
             .iter()
             .filter(|(k, _)| !s3_storage_options::S3_OPTS.contains(&k.as_str()))
@@ -246,19 +254,20 @@ impl S3StorageOptions {
 
         // Copy web identity values provided in options but not the environment into the environment
         // to get picked up by the `from_k8s_env` call in `get_web_identity_provider`.
-        Self::ensure_env_var(&options, s3_storage_options::AWS_REGION);
-        Self::ensure_env_var(&options, s3_storage_options::AWS_ACCESS_KEY_ID);
-        Self::ensure_env_var(&options, s3_storage_options::AWS_SECRET_ACCESS_KEY);
-        Self::ensure_env_var(&options, s3_storage_options::AWS_SESSION_TOKEN);
-        Self::ensure_env_var(&options, s3_storage_options::AWS_WEB_IDENTITY_TOKEN_FILE);
-        Self::ensure_env_var(&options, s3_storage_options::AWS_ROLE_ARN);
-        Self::ensure_env_var(&options, s3_storage_options::AWS_ROLE_SESSION_NAME);
+        Self::ensure_env_var(options, s3_storage_options::AWS_REGION);
+        Self::ensure_env_var(options, s3_storage_options::AWS_PROFILE);
+        Self::ensure_env_var(options, s3_storage_options::AWS_ACCESS_KEY_ID);
+        Self::ensure_env_var(options, s3_storage_options::AWS_SECRET_ACCESS_KEY);
+        Self::ensure_env_var(options, s3_storage_options::AWS_SESSION_TOKEN);
+        Self::ensure_env_var(options, s3_storage_options::AWS_WEB_IDENTITY_TOKEN_FILE);
+        Self::ensure_env_var(options, s3_storage_options::AWS_ROLE_ARN);
+        Self::ensure_env_var(options, s3_storage_options::AWS_ROLE_SESSION_NAME);
 
-        let endpoint_url = str_option(&options, s3_storage_options::AWS_ENDPOINT_URL);
+        let endpoint_url = str_option(options, s3_storage_options::AWS_ENDPOINT_URL);
         let region = if let Some(endpoint_url) = endpoint_url.as_ref() {
             Region::Custom {
                 name: Self::str_or_default(
-                    &options,
+                    options,
                     s3_storage_options::AWS_REGION,
                     "custom".to_string(),
                 ),
@@ -267,34 +276,47 @@ impl S3StorageOptions {
         } else {
             Region::default()
         };
+        let profile = str_option(options, s3_storage_options::AWS_PROFILE);
 
         let s3_pool_idle_timeout = Self::u64_or_default(
-            &options,
+            options,
             s3_storage_options::AWS_S3_POOL_IDLE_TIMEOUT_SECONDS,
             15,
         );
         let sts_pool_idle_timeout = Self::u64_or_default(
-            &options,
+            options,
             s3_storage_options::AWS_STS_POOL_IDLE_TIMEOUT_SECONDS,
             10,
         );
 
         let s3_get_internal_server_error_retries = Self::u64_or_default(
-            &options,
+            options,
             s3_storage_options::AWS_S3_GET_INTERNAL_SERVER_ERROR_RETRIES,
             10,
         ) as usize;
 
+        let virtual_hosted_style_request: bool =
+            str_option(options, s3_storage_options::AWS_S3_ADDRESSING_STYLE)
+                .map(|addressing_style| addressing_style == "virtual")
+                .unwrap_or(false);
+
+        let allow_unsafe_rename =
+            str_option(options, s3_storage_options::AWS_S3_ALLOW_UNSAFE_RENAME)
+                .map(|val| str_is_truthy(&val))
+                .unwrap_or(false);
+
         Self {
             endpoint_url,
             region,
-            aws_access_key_id: str_option(&options, s3_storage_options::AWS_ACCESS_KEY_ID),
-            aws_secret_access_key: str_option(&options, s3_storage_options::AWS_SECRET_ACCESS_KEY),
-            aws_session_token: str_option(&options, s3_storage_options::AWS_SESSION_TOKEN),
-            locking_provider: str_option(&options, s3_storage_options::AWS_S3_LOCKING_PROVIDER),
-            assume_role_arn: str_option(&options, s3_storage_options::AWS_S3_ASSUME_ROLE_ARN),
+            profile,
+            aws_access_key_id: str_option(options, s3_storage_options::AWS_ACCESS_KEY_ID),
+            aws_secret_access_key: str_option(options, s3_storage_options::AWS_SECRET_ACCESS_KEY),
+            aws_session_token: str_option(options, s3_storage_options::AWS_SESSION_TOKEN),
+            virtual_hosted_style_request,
+            locking_provider: str_option(options, s3_storage_options::AWS_S3_LOCKING_PROVIDER),
+            assume_role_arn: str_option(options, s3_storage_options::AWS_S3_ASSUME_ROLE_ARN),
             assume_role_session_name: str_option(
-                &options,
+                options,
                 s3_storage_options::AWS_S3_ROLE_SESSION_NAME,
             ),
             use_web_identity: std::env::var(s3_storage_options::AWS_WEB_IDENTITY_TOKEN_FILE)
@@ -302,6 +324,7 @@ impl S3StorageOptions {
             s3_pool_idle_timeout: Duration::from_secs(s3_pool_idle_timeout),
             sts_pool_idle_timeout: Duration::from_secs(sts_pool_idle_timeout),
             s3_get_internal_server_error_retries,
+            allow_unsafe_rename,
             extra_opts,
         }
     }
@@ -328,7 +351,7 @@ impl S3StorageOptions {
 impl Default for S3StorageOptions {
     /// Creates an instance of S3StorageOptions from environment variables.
     fn default() -> S3StorageOptions {
-        Self::from_map(HashMap::new())
+        Self::from_map(&HashMap::new())
     }
 }
 
@@ -343,6 +366,8 @@ fn get_web_identity_provider() -> Result<AutoRefreshingProvider<WebIdentityProvi
 pub struct S3StorageBackend {
     inner: Arc<DynObjectStore>,
     s3_lock_client: Option<S3LockClient>,
+    /// Whether allowed to performance rename_if_not_exist as rename
+    allow_unsafe_rename: bool,
 }
 
 impl std::fmt::Display for S3StorageBackend {
@@ -378,6 +403,7 @@ impl S3StorageBackend {
         Ok(Self {
             inner: storage,
             s3_lock_client,
+            allow_unsafe_rename: options.allow_unsafe_rename,
         })
     }
 
@@ -390,6 +416,7 @@ impl S3StorageBackend {
         Self {
             inner: storage,
             s3_lock_client,
+            allow_unsafe_rename: false, // Doesn't matter if you provided a lock client
         }
     }
 
@@ -456,11 +483,14 @@ impl ObjectStore for S3StorageBackend {
     }
 
     async fn rename_if_not_exists(&self, from: &Path, to: &Path) -> ObjectStoreResult<()> {
-        let lock_client = match self.s3_lock_client {
-            Some(ref lock_client) => lock_client,
-            None => return Err(S3LockError::LockClientRequired.into()),
-        };
-        lock_client.rename_with_lock(self, from, to).await?;
+        if let Some(lock_client) = &self.s3_lock_client {
+            lock_client.rename_with_lock(self, from, to).await?;
+        } else if self.allow_unsafe_rename {
+            self.inner.rename(from, to).await?;
+        } else {
+            return Err(S3LockError::LockClientRequired.into());
+        }
+
         Ok(())
     }
 
@@ -506,7 +536,6 @@ fn try_create_lock_client(options: &S3StorageOptions) -> Result<Option<S3LockCli
 }
 
 #[cfg(test)]
-#[cfg(not(feature = "integration_test"))]
 mod tests {
     use super::*;
 
@@ -518,6 +547,12 @@ mod tests {
     fn storage_options_default_test() {
         std::env::set_var(s3_storage_options::AWS_ENDPOINT_URL, "http://localhost");
         std::env::set_var(s3_storage_options::AWS_REGION, "us-west-1");
+        std::env::set_var(s3_storage_options::AWS_PROFILE, "default");
+        std::env::set_var(s3_storage_options::AWS_ACCESS_KEY_ID, "default_key_id");
+        std::env::set_var(
+            s3_storage_options::AWS_SECRET_ACCESS_KEY,
+            "default_secret_key",
+        );
         std::env::set_var(s3_storage_options::AWS_S3_LOCKING_PROVIDER, "dynamodb");
         std::env::set_var(
             s3_storage_options::AWS_S3_ASSUME_ROLE_ARN,
@@ -541,9 +576,11 @@ mod tests {
                     name: "us-west-1".to_string(),
                     endpoint: "http://localhost".to_string()
                 },
-                aws_access_key_id: Some("test".to_string()),
-                aws_secret_access_key: Some("test".to_string()),
+                profile: Some("default".to_string()),
+                aws_access_key_id: Some("default_key_id".to_string()),
+                aws_secret_access_key: Some("default_secret_key".to_string()),
                 aws_session_token: None,
+                virtual_hosted_style_request: false,
                 assume_role_arn: Some("arn:aws:iam::123456789012:role/some_role".to_string()),
                 assume_role_session_name: Some("session_name".to_string()),
                 use_web_identity: true,
@@ -552,6 +589,7 @@ mod tests {
                 sts_pool_idle_timeout: Duration::from_secs(10),
                 s3_get_internal_server_error_retries: 10,
                 extra_opts: HashMap::new(),
+                allow_unsafe_rename: false,
             },
             options
         );
@@ -561,10 +599,10 @@ mod tests {
     #[serial]
     fn storage_options_with_only_region_and_credentials() {
         std::env::remove_var(s3_storage_options::AWS_ENDPOINT_URL);
-        let options = S3StorageOptions::from_map(hashmap! {
+        let options = S3StorageOptions::from_map(&hashmap! {
             s3_storage_options::AWS_REGION.to_string() => "eu-west-1".to_string(),
             s3_storage_options::AWS_ACCESS_KEY_ID.to_string() => "test".to_string(),
-            s3_storage_options::AWS_SECRET_ACCESS_KEY.to_string() => "test".to_string(),
+            s3_storage_options::AWS_SECRET_ACCESS_KEY.to_string() => "test_secret".to_string(),
         });
 
         assert_eq!(
@@ -572,7 +610,7 @@ mod tests {
                 endpoint_url: None,
                 region: Region::default(),
                 aws_access_key_id: Some("test".to_string()),
-                aws_secret_access_key: Some("test".to_string()),
+                aws_secret_access_key: Some("test_secret".to_string()),
                 ..Default::default()
             },
             options
@@ -582,9 +620,11 @@ mod tests {
     #[test]
     #[serial]
     fn storage_options_from_map_test() {
-        let options = S3StorageOptions::from_map(hashmap! {
+        let options = S3StorageOptions::from_map(&hashmap! {
             s3_storage_options::AWS_ENDPOINT_URL.to_string() => "http://localhost:1234".to_string(),
             s3_storage_options::AWS_REGION.to_string() => "us-west-2".to_string(),
+            s3_storage_options::AWS_PROFILE.to_string() => "default".to_string(),
+            s3_storage_options::AWS_S3_ADDRESSING_STYLE.to_string() => "virtual".to_string(),
             s3_storage_options::AWS_S3_LOCKING_PROVIDER.to_string() => "another_locking_provider".to_string(),
             s3_storage_options::AWS_S3_ASSUME_ROLE_ARN.to_string() => "arn:aws:iam::123456789012:role/another_role".to_string(),
             s3_storage_options::AWS_S3_ROLE_SESSION_NAME.to_string() => "another_session_name".to_string(),
@@ -592,6 +632,8 @@ mod tests {
             s3_storage_options::AWS_S3_POOL_IDLE_TIMEOUT_SECONDS.to_string() => "1".to_string(),
             s3_storage_options::AWS_STS_POOL_IDLE_TIMEOUT_SECONDS.to_string() => "2".to_string(),
             s3_storage_options::AWS_S3_GET_INTERNAL_SERVER_ERROR_RETRIES.to_string() => "3".to_string(),
+            s3_storage_options::AWS_ACCESS_KEY_ID.to_string() => "test_id".to_string(),
+            s3_storage_options::AWS_SECRET_ACCESS_KEY.to_string() => "test_secret".to_string(),
         });
 
         assert_eq!(
@@ -601,9 +643,11 @@ mod tests {
                     name: "us-west-2".to_string(),
                     endpoint: "http://localhost:1234".to_string()
                 },
-                aws_access_key_id: Some("test".to_string()),
-                aws_secret_access_key: Some("test".to_string()),
+                profile: Some("default".to_string()),
+                aws_access_key_id: Some("test_id".to_string()),
+                aws_secret_access_key: Some("test_secret".to_string()),
                 aws_session_token: None,
+                virtual_hosted_style_request: true,
                 assume_role_arn: Some("arn:aws:iam::123456789012:role/another_role".to_string()),
                 assume_role_session_name: Some("another_session_name".to_string()),
                 use_web_identity: true,
@@ -611,7 +655,10 @@ mod tests {
                 s3_pool_idle_timeout: Duration::from_secs(1),
                 sts_pool_idle_timeout: Duration::from_secs(2),
                 s3_get_internal_server_error_retries: 3,
-                extra_opts: HashMap::new(),
+                extra_opts: hashmap! {
+                    s3_storage_options::AWS_S3_ADDRESSING_STYLE.to_string() => "virtual".to_string()
+                },
+                allow_unsafe_rename: false,
             },
             options
         );
@@ -622,6 +669,12 @@ mod tests {
     fn storage_options_mixed_test() {
         std::env::set_var(s3_storage_options::AWS_ENDPOINT_URL, "http://localhost");
         std::env::set_var(s3_storage_options::AWS_REGION, "us-west-1");
+        std::env::set_var(s3_storage_options::AWS_PROFILE, "default");
+        std::env::set_var(s3_storage_options::AWS_ACCESS_KEY_ID, "wrong_key_id");
+        std::env::set_var(
+            s3_storage_options::AWS_SECRET_ACCESS_KEY,
+            "wrong_secret_key",
+        );
         std::env::set_var(s3_storage_options::AWS_S3_LOCKING_PROVIDER, "dynamodb");
         std::env::set_var(
             s3_storage_options::AWS_S3_ASSUME_ROLE_ARN,
@@ -639,7 +692,9 @@ mod tests {
             s3_storage_options::AWS_S3_GET_INTERNAL_SERVER_ERROR_RETRIES,
             "3",
         );
-        let options = S3StorageOptions::from_map(hashmap! {
+        let options = S3StorageOptions::from_map(&hashmap! {
+            s3_storage_options::AWS_ACCESS_KEY_ID.to_string() => "test_id_mixed".to_string(),
+            s3_storage_options::AWS_SECRET_ACCESS_KEY.to_string() => "test_secret_mixed".to_string(),
             s3_storage_options::AWS_REGION.to_string() => "us-west-2".to_string(),
             "DYNAMO_LOCK_PARTITION_KEY_VALUE".to_string() => "my_lock".to_string(),
             "AWS_S3_GET_INTERNAL_SERVER_ERROR_RETRIES".to_string() => "3".to_string(),
@@ -652,9 +707,11 @@ mod tests {
                     name: "us-west-2".to_string(),
                     endpoint: "http://localhost".to_string()
                 },
-                aws_access_key_id: Some("test".to_string()),
-                aws_secret_access_key: Some("test".to_string()),
+                profile: Some("default".to_string()),
+                aws_access_key_id: Some("test_id_mixed".to_string()),
+                aws_secret_access_key: Some("test_secret_mixed".to_string()),
                 aws_session_token: None,
+                virtual_hosted_style_request: false,
                 assume_role_arn: Some("arn:aws:iam::123456789012:role/some_role".to_string()),
                 assume_role_session_name: Some("session_name".to_string()),
                 use_web_identity: true,
@@ -665,6 +722,7 @@ mod tests {
                 extra_opts: hashmap! {
                     "DYNAMO_LOCK_PARTITION_KEY_VALUE".to_string() => "my_lock".to_string(),
                 },
+                allow_unsafe_rename: false,
             },
             options
         );
@@ -672,7 +730,7 @@ mod tests {
     #[test]
     #[serial]
     fn storage_options_web_identity_test() {
-        let _options = S3StorageOptions::from_map(hashmap! {
+        let _options = S3StorageOptions::from_map(&hashmap! {
             s3_storage_options::AWS_REGION.to_string() => "eu-west-1".to_string(),
             s3_storage_options::AWS_WEB_IDENTITY_TOKEN_FILE.to_string() => "web_identity_token_file".to_string(),
             s3_storage_options::AWS_ROLE_ARN.to_string() => "arn:aws:iam::123456789012:role/web_identity_role".to_string(),
