@@ -183,7 +183,7 @@ impl JsonWriter {
         partition_columns: Option<Vec<String>>,
         storage_options: Option<HashMap<String, String>>,
     ) -> Result<Self, DeltaTableError> {
-        let storage = DeltaTableBuilder::from_uri(&table_uri)
+        let storage = DeltaTableBuilder::from_uri(table_uri)
             .with_storage_options(storage_options.unwrap_or_default())
             .build_storage()?;
 
@@ -339,16 +339,20 @@ impl DeltaWriter<Vec<Value>> for JsonWriter {
         }
 
         if !partial_writes.is_empty() {
-            let sample = partial_writes.first().map(|t| t.to_owned());
-            if let Some((_, e)) = sample {
-                return Err(DeltaWriterError::PartialParquetWrite {
-                    skipped_values: partial_writes,
-                    sample_error: e,
-                }
-                .into());
-            } else {
-                unreachable!()
+            return Err(DeltaWriterError::PartialParquetWrite {
+                sample_error: match &partial_writes[0].1 {
+                    ParquetError::General(msg) => ParquetError::General(msg.to_owned()),
+                    ParquetError::ArrowError(msg) => ParquetError::ArrowError(msg.to_owned()),
+                    ParquetError::EOF(msg) => ParquetError::EOF(msg.to_owned()),
+                    ParquetError::External(err) => ParquetError::General(err.to_string()),
+                    ParquetError::IndexOutOfBound(u, v) => {
+                        ParquetError::IndexOutOfBound(u.to_owned(), v.to_owned())
+                    }
+                    ParquetError::NYI(msg) => ParquetError::NYI(msg.to_owned()),
+                },
+                skipped_values: partial_writes,
             }
+            .into());
         }
 
         Ok(())
@@ -437,11 +441,15 @@ fn extract_partition_values(
 
 #[cfg(test)]
 mod tests {
+    use super::*;
+    use crate::arrow::array::Int32Array;
+    use crate::arrow::datatypes::{
+        DataType as ArrowDataType, Field as ArrowField, Schema as ArrowSchema,
+    };
     use crate::writer::test_utils::get_delta_schema;
     use crate::writer::DeltaWriter;
     use crate::writer::JsonWriter;
     use crate::Schema;
-    use arrow::datatypes::Schema as ArrowSchema;
     use parquet::file::reader::FileReader;
     use parquet::file::serialized_reader::SerializedFileReader;
     use std::fs::File;
@@ -488,4 +496,43 @@ mod tests {
             .collect::<Vec<String>>();
         assert_eq!(columns, vec!["id".to_string(), "value".to_string()]);
     }
+
+    #[test]
+    fn test_extract_partition_values() {
+        let record_batch = RecordBatch::try_new(
+            Arc::new(ArrowSchema::new(vec![
+                ArrowField::new("col1", ArrowDataType::Int32, false),
+                ArrowField::new("col2", ArrowDataType::Int32, false),
+                ArrowField::new("col3", ArrowDataType::Int32, true),
+            ])),
+            vec![
+                Arc::new(Int32Array::from(vec![1, 2])),
+                Arc::new(Int32Array::from(vec![2, 1])),
+                Arc::new(Int32Array::from(vec![None, None])),
+            ],
+        )
+        .unwrap();
+
+        assert_eq!(
+            extract_partition_values(
+                &[
+                    String::from("col1"),
+                    String::from("col2"),
+                    String::from("col3")
+                ],
+                &record_batch
+            )
+            .unwrap(),
+            HashMap::from([
+                (String::from("col1"), Some(String::from("1"))),
+                (String::from("col2"), Some(String::from("2"))),
+                (String::from("col3"), None),
+            ])
+        );
+        assert_eq!(
+            extract_partition_values(&[String::from("col1")], &record_batch).unwrap(),
+            HashMap::from([(String::from("col1"), Some(String::from("1"))),])
+        );
+        assert!(extract_partition_values(&[String::from("col4")], &record_batch).is_err())
+    }
 }
