@@ -11,12 +11,10 @@ use arrow::{
     },
     buffer::MutableBuffer,
 };
-use parquet::{
-    errors::ParquetError,
-    file::{metadata::RowGroupMetaData, statistics::Statistics},
-    schema::types::{ColumnDescriptor, SchemaDescriptor},
-};
-use parquet_format::FileMetaData;
+use parquet::errors::ParquetError;
+use parquet::file::{metadata::RowGroupMetaData, statistics::Statistics};
+use parquet::format::FileMetaData;
+use parquet::schema::types::{ColumnDescriptor, SchemaDescriptor};
 use serde_json::{Number, Value};
 use std::collections::HashMap;
 use std::sync::Arc;
@@ -31,7 +29,7 @@ pub type MinAndMaxValues = (
 pub(crate) fn apply_null_counts(
     array: &StructArray,
     null_counts: &mut HashMap<String, ColumnCountStat>,
-    nest_level: i32,
+    _nest_level: i32,
 ) {
     let fields = match array.data_type() {
         DataType::Struct(fields) => fields,
@@ -40,7 +38,7 @@ pub(crate) fn apply_null_counts(
 
     array
         .columns()
-        .into_iter()
+        .iter()
         .zip(fields)
         .for_each(|(column, field)| {
             let key = field.name().to_owned();
@@ -54,7 +52,7 @@ pub(crate) fn apply_null_counts(
 
                     match col_struct {
                         ColumnCountStat::Column(map) => {
-                            apply_null_counts(as_struct_array(column), map, nest_level + 1);
+                            apply_null_counts(as_struct_array(column), map, _nest_level + 1);
                         }
                         _ => unreachable!(),
                     }
@@ -95,8 +93,7 @@ pub(crate) fn create_add(
         null_count: null_counts,
     };
 
-    let stats_string = serde_json::to_string(&stats)
-        .or(Err(DeltaWriterError::StatsSerializationFailed { stats }))?;
+    let stats_string = serde_json::to_string(&stats)?;
 
     // Determine the modification timestamp to include in the add action - milliseconds since epoch
     // Err should be impossible in this case since `SystemTime::now()` is always greater than `UNIX_EPOCH`
@@ -317,8 +314,10 @@ fn min_and_max_from_parquet_statistics(
 
             match column_descr.logical_type().as_ref() {
                 Some(LogicalType::Timestamp { unit, .. }) => {
-                    let min = min.map(|n| Value::String(timestamp_to_delta_stats_string(n, unit)));
-                    let max = max.map(|n| Value::String(timestamp_to_delta_stats_string(n, unit)));
+                    let min = min
+                        .and_then(|n| timestamp_to_delta_stats_string(n, unit).map(Value::String));
+                    let max = max
+                        .and_then(|n| timestamp_to_delta_stats_string(n, unit).map(Value::String));
 
                     Ok((min, max))
                 }
