@@ -1,7 +1,7 @@
 use std::collections::HashMap;
 use std::sync::Arc;
 
-use crate::utils::{delete_dir, wait_for_future, walk_tree};
+use crate::utils::{delete_dir, rt, walk_tree};
 use crate::PyDeltaTableError;
 
 use deltalake::storage::{DynObjectStore, ListResult, MultipartId, ObjectStoreError, Path};
@@ -9,24 +9,41 @@ use deltalake::DeltaTableBuilder;
 use pyo3::exceptions::{PyIOError, PyNotImplementedError, PyValueError};
 use pyo3::prelude::*;
 use pyo3::types::{IntoPyDict, PyBytes};
+use serde::{Deserialize, Serialize};
 use tokio::io::{AsyncWrite, AsyncWriteExt};
+use tokio::runtime::Runtime;
+
+#[derive(Debug, Clone, Serialize, Deserialize)]
+pub(crate) struct FsConfig {
+    pub(crate) root_url: String,
+    pub(crate) options: HashMap<String, String>,
+}
 
 #[pyclass(subclass)]
 #[derive(Debug, Clone)]
 pub struct DeltaFileSystemHandler {
     pub(crate) inner: Arc<DynObjectStore>,
+    pub(crate) rt: Arc<Runtime>,
+    pub(crate) config: FsConfig,
 }
 
 #[pymethods]
 impl DeltaFileSystemHandler {
     #[new]
-    #[args(options = "None")]
+    #[pyo3(signature = (table_uri, options = None))]
     fn new(table_uri: &str, options: Option<HashMap<String, String>>) -> PyResult<Self> {
         let storage = DeltaTableBuilder::from_uri(table_uri)
-            .with_storage_options(options.unwrap_or_default())
+            .with_storage_options(options.clone().unwrap_or_default())
             .build_storage()
             .map_err(PyDeltaTableError::from_raw)?;
-        Ok(Self { inner: storage })
+        Ok(Self {
+            inner: storage,
+            rt: Arc::new(rt()?),
+            config: FsConfig {
+                root_url: table_uri.into(),
+                options: options.unwrap_or_default(),
+            },
+        })
     }
 
     fn get_type_name(&self) -> String {
@@ -36,13 +53,14 @@ impl DeltaFileSystemHandler {
     fn normalize_path(&self, path: String) -> PyResult<String> {
         let suffix = if path.ends_with('/') { "/" } else { "" };
         let path = Path::parse(path).unwrap();
-        Ok(format!("{}{}", path, suffix))
+        Ok(format!("{path}{suffix}"))
     }
 
-    fn copy_file(&self, src: String, dest: String, py: Python) -> PyResult<()> {
+    fn copy_file(&self, src: String, dest: String) -> PyResult<()> {
         let from_path = Path::from(src);
         let to_path = Path::from(dest);
-        wait_for_future(py, self.inner.copy(&from_path, &to_path))
+        self.rt
+            .block_on(self.inner.copy(&from_path, &to_path))
             .map_err(PyDeltaTableError::from_object_store)?;
         Ok(())
     }
@@ -52,41 +70,45 @@ impl DeltaFileSystemHandler {
         Ok(())
     }
 
-    fn delete_dir(&self, path: String, py: Python) -> PyResult<()> {
+    fn delete_dir(&self, path: String) -> PyResult<()> {
         let path = Path::from(path);
-        wait_for_future(py, delete_dir(self.inner.as_ref(), &path))
+        self.rt
+            .block_on(delete_dir(self.inner.as_ref(), &path))
             .map_err(PyDeltaTableError::from_object_store)?;
         Ok(())
     }
 
-    fn delete_file(&self, path: String, py: Python) -> PyResult<()> {
+    fn delete_file(&self, path: String) -> PyResult<()> {
         let path = Path::from(path);
-        wait_for_future(py, self.inner.delete(&path))
+        self.rt
+            .block_on(self.inner.delete(&path))
             .map_err(PyDeltaTableError::from_object_store)?;
         Ok(())
     }
 
     fn equals(&self, other: &DeltaFileSystemHandler) -> PyResult<bool> {
-        Ok(format!("{:?}", self) == format!("{:?}", other))
+        Ok(format!("{self:?}") == format!("{other:?}"))
     }
 
     fn get_file_info<'py>(&self, paths: Vec<String>, py: Python<'py>) -> PyResult<Vec<&'py PyAny>> {
         let fs = PyModule::import(py, "pyarrow.fs")?;
         let file_types = fs.getattr("FileType")?;
 
-        let to_file_info = |loc: String, type_: &PyAny, kwargs: HashMap<&str, i64>| {
+        let to_file_info = |loc: &str, type_: &PyAny, kwargs: &HashMap<&str, i64>| {
             fs.call_method("FileInfo", (loc, type_), Some(kwargs.into_py_dict(py)))
         };
 
         let mut infos = Vec::new();
         for file_path in paths {
             let path = Path::from(file_path);
-            let listed = wait_for_future(py, self.inner.list_with_delimiter(Some(&path)))
+            let listed = self
+                .rt
+                .block_on(self.inner.list_with_delimiter(Some(&path)))
                 .map_err(PyDeltaTableError::from_object_store)?;
 
             // TODO is there a better way to figure out if we are in a directory?
             if listed.objects.is_empty() && listed.common_prefixes.is_empty() {
-                let maybe_meta = wait_for_future(py, self.inner.head(&path));
+                let maybe_meta = self.rt.block_on(self.inner.head(&path));
                 match maybe_meta {
                     Ok(meta) => {
                         let kwargs = HashMap::from([
@@ -94,16 +116,16 @@ impl DeltaFileSystemHandler {
                             ("mtime_ns", meta.last_modified.timestamp_nanos()),
                         ]);
                         infos.push(to_file_info(
-                            meta.location.to_string(),
+                            meta.location.as_ref(),
                             file_types.getattr("File")?,
-                            kwargs,
+                            &kwargs,
                         )?);
                     }
                     Err(ObjectStoreError::NotFound { .. }) => {
                         infos.push(to_file_info(
-                            path.to_string(),
+                            path.as_ref(),
                             file_types.getattr("NotFound")?,
-                            HashMap::new(),
+                            &HashMap::new(),
                         )?);
                     }
                     Err(err) => {
@@ -112,9 +134,9 @@ impl DeltaFileSystemHandler {
                 }
             } else {
                 infos.push(to_file_info(
-                    path.to_string(),
+                    path.as_ref(),
                     file_types.getattr("Directory")?,
-                    HashMap::new(),
+                    &HashMap::new(),
                 )?);
             }
         }
@@ -122,7 +144,7 @@ impl DeltaFileSystemHandler {
         Ok(infos)
     }
 
-    #[args(allow_not_found = "false", recursive = "false")]
+    #[pyo3(signature = (base_dir, allow_not_found = false, recursive = false))]
     fn get_file_info_selector<'py>(
         &self,
         base_dir: String,
@@ -138,22 +160,24 @@ impl DeltaFileSystemHandler {
         };
 
         let path = Path::from(base_dir);
-        let list_result =
-            match wait_for_future(py, walk_tree(self.inner.clone(), &path, recursive)) {
-                Ok(res) => Ok(res),
-                Err(ObjectStoreError::NotFound { path, source }) => {
-                    if allow_not_found {
-                        Ok(ListResult {
-                            common_prefixes: vec![],
-                            objects: vec![],
-                        })
-                    } else {
-                        Err(ObjectStoreError::NotFound { path, source })
-                    }
+        let list_result = match self
+            .rt
+            .block_on(walk_tree(self.inner.clone(), &path, recursive))
+        {
+            Ok(res) => Ok(res),
+            Err(ObjectStoreError::NotFound { path, source }) => {
+                if allow_not_found {
+                    Ok(ListResult {
+                        common_prefixes: vec![],
+                        objects: vec![],
+                    })
+                } else {
+                    Err(ObjectStoreError::NotFound { path, source })
                 }
-                Err(err) => Err(err),
             }
-            .map_err(PyDeltaTableError::from_object_store)?;
+            Err(err) => Err(err),
+        }
+        .map_err(PyDeltaTableError::from_object_store)?;
 
         let mut infos = vec![];
         infos.extend(
@@ -190,34 +214,53 @@ impl DeltaFileSystemHandler {
         Ok(infos)
     }
 
-    fn move_file(&self, src: String, dest: String, py: Python) -> PyResult<()> {
+    fn move_file(&self, src: String, dest: String) -> PyResult<()> {
         let from_path = Path::from(src);
         let to_path = Path::from(dest);
         // TODO check the if not exists semantics
-        wait_for_future(py, self.inner.rename(&from_path, &to_path))
+        self.rt
+            .block_on(self.inner.rename(&from_path, &to_path))
             .map_err(PyDeltaTableError::from_object_store)?;
         Ok(())
     }
 
-    fn open_input_file(&self, path: String, py: Python) -> PyResult<ObjectInputFile> {
+    fn open_input_file(&self, path: String) -> PyResult<ObjectInputFile> {
         let path = Path::from(path);
-        let file = wait_for_future(py, ObjectInputFile::try_new(self.inner.clone(), path))
+        let file = self
+            .rt
+            .block_on(ObjectInputFile::try_new(
+                Arc::clone(&self.rt),
+                self.inner.clone(),
+                path,
+            ))
             .map_err(PyDeltaTableError::from_object_store)?;
         Ok(file)
     }
 
-    #[args(metadata = "None")]
+    #[pyo3(signature = (path, metadata = None))]
     fn open_output_stream(
         &self,
         path: String,
         #[allow(unused)] metadata: Option<HashMap<String, String>>,
-        py: Python,
     ) -> PyResult<ObjectOutputStream> {
         let path = Path::from(path);
-        let file = wait_for_future(py, ObjectOutputStream::try_new(self.inner.clone(), path))
+        let file = self
+            .rt
+            .block_on(ObjectOutputStream::try_new(
+                Arc::clone(&self.rt),
+                self.inner.clone(),
+                path,
+            ))
             .map_err(PyDeltaTableError::from_object_store)?;
         Ok(file)
     }
+
+    pub fn __getnewargs__(&self) -> PyResult<(String, Option<HashMap<String, String>>)> {
+        Ok((
+            self.config.root_url.clone(),
+            Some(self.config.options.clone()),
+        ))
+    }
 }
 
 // TODO the C++ implementation track an internal lock on all random access files, DO we need this here?
@@ -226,6 +269,7 @@ impl DeltaFileSystemHandler {
 #[derive(Debug, Clone)]
 pub struct ObjectInputFile {
     store: Arc<DynObjectStore>,
+    rt: Arc<Runtime>,
     path: Path,
     content_length: i64,
     #[pyo3(get)]
@@ -236,7 +280,11 @@ pub struct ObjectInputFile {
 }
 
 impl ObjectInputFile {
-    pub async fn try_new(store: Arc<DynObjectStore>, path: Path) -> Result<Self, ObjectStoreError> {
+    pub async fn try_new(
+        rt: Arc<Runtime>,
+        store: Arc<DynObjectStore>,
+        path: Path,
+    ) -> Result<Self, ObjectStoreError> {
         // Issue a HEAD Object to get the content-length and ensure any
         // errors (e.g. file not found) don't wait until the first read() call.
         let meta = store.head(&path).await?;
@@ -245,6 +293,7 @@ impl ObjectInputFile {
         // https://github.com/apache/arrow/blob/f184255cbb9bf911ea2a04910f711e1a924b12b8/cpp/src/arrow/filesystem/s3fs.cc#L1083
         Ok(Self {
             store,
+            rt,
             path,
             content_length,
             closed: false,
@@ -264,14 +313,12 @@ impl ObjectInputFile {
     fn check_position(&self, position: i64, action: &str) -> PyResult<()> {
         if position < 0 {
             return Err(PyIOError::new_err(format!(
-                "Cannot {} for negative position.",
-                action
+                "Cannot {action} for negative position."
             )));
         }
         if position > self.content_length {
             return Err(PyIOError::new_err(format!(
-                "Cannot {} past end of file.",
-                action
+                "Cannot {action} past end of file."
             )));
         }
         Ok(())
@@ -311,7 +358,7 @@ impl ObjectInputFile {
         Ok(self.content_length)
     }
 
-    #[args(whence = "0")]
+    #[pyo3(signature = (offset, whence = 0))]
     fn seek(&mut self, offset: i64, whence: i64) -> PyResult<i64> {
         self.check_closed()?;
         self.check_position(offset, "seek")?;
@@ -326,7 +373,7 @@ impl ObjectInputFile {
             }
             // reference is  end of the stream; offset is usually negative
             2 => {
-                self.pos = self.content_length as i64 + offset;
+                self.pos = self.content_length + offset;
             }
             _ => {
                 return Err(PyValueError::new_err(
@@ -337,8 +384,8 @@ impl ObjectInputFile {
         Ok(self.pos)
     }
 
-    #[args(nbytes = "None")]
-    fn read<'py>(&mut self, nbytes: Option<i64>, py: Python<'py>) -> PyResult<&'py PyBytes> {
+    #[pyo3(signature = (nbytes = None))]
+    fn read(&mut self, nbytes: Option<i64>) -> PyResult<Py<PyBytes>> {
         self.check_closed()?;
         let range = match nbytes {
             Some(len) => {
@@ -355,14 +402,14 @@ impl ObjectInputFile {
         };
         let nbytes = (range.end - range.start) as i64;
         self.pos += nbytes;
-        let obj = if nbytes > 0 {
-            wait_for_future(py, self.store.get_range(&self.path, range))
+        let data = if nbytes > 0 {
+            self.rt
+                .block_on(self.store.get_range(&self.path, range))
                 .map_err(PyDeltaTableError::from_object_store)?
-                .to_vec()
         } else {
-            Vec::new()
+            "".into()
         };
-        Ok(PyBytes::new(py, &obj))
+        Python::with_gil(|py| Ok(PyBytes::new(py, data.as_ref()).into_py(py)))
     }
 
     fn fileno(&self) -> PyResult<()> {
@@ -389,6 +436,7 @@ impl ObjectInputFile {
 #[pyclass(weakref)]
 pub struct ObjectOutputStream {
     store: Arc<DynObjectStore>,
+    rt: Arc<Runtime>,
     path: Path,
     writer: Box<dyn AsyncWrite + Send + Unpin>,
     multipart_id: MultipartId,
@@ -400,10 +448,15 @@ pub struct ObjectOutputStream {
 }
 
 impl ObjectOutputStream {
-    pub async fn try_new(store: Arc<DynObjectStore>, path: Path) -> Result<Self, ObjectStoreError> {
-        let (multipart_id, writer) = store.put_multipart(&path).await.unwrap();
+    pub async fn try_new(
+        rt: Arc<Runtime>,
+        store: Arc<DynObjectStore>,
+        path: Path,
+    ) -> Result<Self, ObjectStoreError> {
+        let (multipart_id, writer) = store.put_multipart(&path).await?;
         Ok(Self {
             store,
+            rt,
             path,
             writer,
             multipart_id,
@@ -424,16 +477,14 @@ impl ObjectOutputStream {
 
 #[pymethods]
 impl ObjectOutputStream {
-    fn close(&mut self, py: Python) -> PyResult<()> {
+    fn close(&mut self) -> PyResult<()> {
         self.closed = true;
-        match wait_for_future(py, self.writer.shutdown()) {
+        match self.rt.block_on(self.writer.shutdown()) {
             Ok(_) => Ok(()),
             Err(err) => {
-                wait_for_future(
-                    py,
-                    self.store.abort_multipart(&self.path, &self.multipart_id),
-                )
-                .map_err(PyDeltaTableError::from_object_store)?;
+                self.rt
+                    .block_on(self.store.abort_multipart(&self.path, &self.multipart_id))
+                    .map_err(PyDeltaTableError::from_object_store)?;
                 Err(PyDeltaTableError::from_io(err))
             }
         }
@@ -465,43 +516,41 @@ impl ObjectOutputStream {
         Err(PyNotImplementedError::new_err("'size' not implemented"))
     }
 
-    #[args(whence = "0")]
-    fn seek(&mut self, _offset: i64, _whence: i64) -> PyResult<i64> {
+    #[allow(unused_variables)]
+    #[pyo3(signature = (offset, whence = 0))]
+    fn seek(&mut self, offset: i64, whence: i64) -> PyResult<i64> {
         self.check_closed()?;
         Err(PyNotImplementedError::new_err("'seek' not implemented"))
     }
 
-    #[args(nbytes = "None")]
-    fn read(&mut self, _nbytes: Option<i64>) -> PyResult<()> {
+    #[allow(unused_variables)]
+    #[pyo3(signature = (nbytes = None))]
+    fn read(&mut self, nbytes: Option<i64>) -> PyResult<()> {
         self.check_closed()?;
         Err(PyNotImplementedError::new_err("'read' not implemented"))
     }
 
-    fn write(&mut self, data: Vec<u8>, py: Python) -> PyResult<i64> {
+    fn write(&mut self, data: &PyBytes) -> PyResult<i64> {
         self.check_closed()?;
-        let len = data.len() as i64;
-        match wait_for_future(py, self.writer.write_all(&data)) {
+        let len = data.as_bytes().len() as i64;
+        match self.rt.block_on(self.writer.write_all(data.as_bytes())) {
             Ok(_) => Ok(len),
             Err(err) => {
-                wait_for_future(
-                    py,
-                    self.store.abort_multipart(&self.path, &self.multipart_id),
-                )
-                .map_err(PyDeltaTableError::from_object_store)?;
+                self.rt
+                    .block_on(self.store.abort_multipart(&self.path, &self.multipart_id))
+                    .map_err(PyDeltaTableError::from_object_store)?;
                 Err(PyDeltaTableError::from_io(err))
             }
         }
     }
 
-    fn flush(&mut self, py: Python) -> PyResult<()> {
-        match wait_for_future(py, self.writer.flush()) {
+    fn flush(&mut self) -> PyResult<()> {
+        match self.rt.block_on(self.writer.flush()) {
             Ok(_) => Ok(()),
             Err(err) => {
-                wait_for_future(
-                    py,
-                    self.store.abort_multipart(&self.path, &self.multipart_id),
-                )
-                .map_err(PyDeltaTableError::from_object_store)?;
+                self.rt
+                    .block_on(self.store.abort_multipart(&self.path, &self.multipart_id))
+                    .map_err(PyDeltaTableError::from_object_store)?;
                 Err(PyDeltaTableError::from_io(err))
             }
         }
