@@ -4,20 +4,22 @@ mod filesystem;
 mod schema;
 mod utils;
 
-use chrono::{DateTime, FixedOffset, Utc};
+use arrow::pyarrow::PyArrowType;
+use chrono::{DateTime, Duration, FixedOffset, Utc};
 use deltalake::action::{
     self, Action, ColumnCountStat, ColumnValueStat, DeltaOperation, SaveMode, Stats,
 };
 use deltalake::arrow::record_batch::RecordBatch;
 use deltalake::arrow::{self, datatypes::Schema as ArrowSchema};
 use deltalake::builder::DeltaTableBuilder;
+use deltalake::checkpoints::create_checkpoint;
 use deltalake::delta_datafusion::DeltaDataChecker;
+use deltalake::operations::vacuum::VacuumBuilder;
 use deltalake::partitions::PartitionFilter;
-use deltalake::DeltaDataTypeLong;
-use deltalake::DeltaDataTypeTimestamp;
-use deltalake::DeltaTableMetaData;
-use deltalake::DeltaTransactionOptions;
-use deltalake::{Invariant, Schema};
+use deltalake::{
+    DeltaDataTypeLong, DeltaDataTypeTimestamp, DeltaTableMetaData, DeltaTransactionOptions,
+    Invariant, Schema,
+};
 use pyo3::create_exception;
 use pyo3::exceptions::PyException;
 use pyo3::exceptions::PyValueError;
@@ -26,9 +28,11 @@ use pyo3::types::PyType;
 use std::collections::HashMap;
 use std::collections::HashSet;
 use std::convert::TryFrom;
+use std::sync::Arc;
 use std::time::SystemTime;
 use std::time::UNIX_EPOCH;
 
+use crate::filesystem::FsConfig;
 use crate::schema::schema_to_pyobject;
 
 create_exception!(deltalake, PyDeltaTableError, PyException);
@@ -46,10 +50,6 @@ impl PyDeltaTableError {
         PyDeltaTableError::new_err(err.to_string())
     }
 
-    fn from_vacuum_error(err: deltalake::vacuum::VacuumError) -> pyo3::PyErr {
-        PyDeltaTableError::new_err(err.to_string())
-    }
-
     fn from_tokio(err: tokio::io::Error) -> pyo3::PyErr {
         PyDeltaTableError::new_err(err.to_string())
     }
@@ -63,7 +63,11 @@ impl PyDeltaTableError {
     }
 
     fn from_chrono(err: chrono::ParseError) -> pyo3::PyErr {
-        PyDeltaTableError::new_err(format!("Parse date and time string failed: {}", err))
+        PyDeltaTableError::new_err(format!("Parse date and time string failed: {err}"))
+    }
+
+    fn from_checkpoint(err: deltalake::checkpoints::CheckpointError) -> pyo3::PyErr {
+        PyDeltaTableError::new_err(err.to_string())
     }
 }
 
@@ -81,6 +85,8 @@ enum PartitionFilterValue<'a> {
 #[pyclass]
 struct RawDeltaTable {
     _table: deltalake::DeltaTable,
+    // storing the config additionally on the table helps us make pickling work.
+    _config: FsConfig,
 }
 
 #[pyclass]
@@ -102,22 +108,36 @@ struct RawDeltaTableMetaData {
 #[pymethods]
 impl RawDeltaTable {
     #[new]
+    #[pyo3(signature = (table_uri, version = None, storage_options = None, without_files = false))]
     fn new(
         table_uri: &str,
         version: Option<deltalake::DeltaDataTypeLong>,
         storage_options: Option<HashMap<String, String>>,
+        without_files: bool,
     ) -> PyResult<Self> {
         let mut builder = deltalake::DeltaTableBuilder::from_uri(table_uri);
+        let options = storage_options.clone().unwrap_or_default();
         if let Some(storage_options) = storage_options {
             builder = builder.with_storage_options(storage_options)
         }
         if let Some(version) = version {
             builder = builder.with_version(version)
         }
+
+        if without_files {
+            builder = builder.without_files()
+        }
+
         let table = rt()?
             .block_on(builder.load())
             .map_err(PyDeltaTableError::from_raw)?;
-        Ok(RawDeltaTable { _table: table })
+        Ok(RawDeltaTable {
+            _table: table,
+            _config: FsConfig {
+                root_url: table_uri.into(),
+                options,
+            },
+        })
     }
 
     #[classmethod]
@@ -215,16 +235,43 @@ impl RawDeltaTable {
         }
     }
 
-    pub fn files(&self) -> PyResult<Vec<String>> {
-        Ok(self
-            ._table
-            .get_files_iter()
-            .map(|f| f.to_string())
-            .collect())
+    pub fn files(
+        &self,
+        partition_filters: Option<Vec<(&str, &str, PartitionFilterValue)>>,
+    ) -> PyResult<Vec<String>> {
+        if let Some(filters) = partition_filters {
+            let filters =
+                convert_partition_filters(filters).map_err(PyDeltaTableError::from_raw)?;
+            Ok(self
+                ._table
+                .get_files_by_partitions(&filters)
+                .map_err(PyDeltaTableError::from_raw)?
+                .into_iter()
+                .map(|p| p.to_string())
+                .collect())
+        } else {
+            Ok(self
+                ._table
+                .get_files_iter()
+                .map(|f| f.to_string())
+                .collect())
+        }
     }
 
-    pub fn file_uris(&self) -> PyResult<Vec<String>> {
-        Ok(self._table.get_file_uris().collect())
+    pub fn file_uris(
+        &self,
+        partition_filters: Option<Vec<(&str, &str, PartitionFilterValue)>>,
+    ) -> PyResult<Vec<String>> {
+        if let Some(filters) = partition_filters {
+            let filters =
+                convert_partition_filters(filters).map_err(PyDeltaTableError::from_raw)?;
+            Ok(self
+                ._table
+                .get_file_uris_by_partitions(&filters)
+                .map_err(PyDeltaTableError::from_raw)?)
+        } else {
+            Ok(self._table.get_file_uris().collect())
+        }
     }
 
     #[getter]
@@ -236,19 +283,26 @@ impl RawDeltaTable {
         schema_to_pyobject(schema, py)
     }
 
-    /// Run the Vacuum command on the Delta Table: list and delete files no longer referenced by the Delta table and are older than the retention threshold.
+    /// Run the Vacuum command on the Delta Table: list and delete files no longer referenced
+    /// by the Delta table and are older than the retention threshold.
+    #[pyo3(signature = (dry_run, retention_hours = None, enforce_retention_duration = true))]
     pub fn vacuum(
         &mut self,
         dry_run: bool,
         retention_hours: Option<u64>,
         enforce_retention_duration: bool,
     ) -> PyResult<Vec<String>> {
-        rt()?
-            .block_on(
-                self._table
-                    .vacuum(retention_hours, dry_run, enforce_retention_duration),
-            )
-            .map_err(PyDeltaTableError::from_vacuum_error)
+        let mut cmd = VacuumBuilder::new(self._table.object_store(), self._table.state.clone())
+            .with_enforce_retention_duration(enforce_retention_duration)
+            .with_dry_run(dry_run);
+        if let Some(retention_period) = retention_hours {
+            cmd = cmd.with_retention_period(Duration::hours(retention_period as i64));
+        }
+        let (table, metrics) = rt()?
+            .block_on(async { cmd.await })
+            .map_err(PyDeltaTableError::from_raw)?;
+        self._table.state = table.state;
+        Ok(metrics.files_deleted)
     }
 
     // Run the History command on the Delta Table: Returns provenance information, including the operation, user, and so on, for each write to a table.
@@ -269,23 +323,22 @@ impl RawDeltaTable {
             .map_err(PyDeltaTableError::from_raw)?;
         serde_json::to_string(
             &<ArrowSchema as TryFrom<&deltalake::Schema>>::try_from(schema)
-                .map_err(PyDeltaTableError::from_arrow)?
-                .to_json(),
+                .map_err(PyDeltaTableError::from_arrow)?,
         )
         .map_err(|_| PyDeltaTableError::new_err("Got invalid table schema"))
     }
 
     pub fn update_incremental(&mut self) -> PyResult<()> {
         rt()?
-            .block_on(self._table.update_incremental())
+            .block_on(self._table.update_incremental(None))
             .map_err(PyDeltaTableError::from_raw)
     }
 
     pub fn dataset_partitions<'py>(
         &mut self,
         py: Python<'py>,
+        schema: PyArrowType<ArrowSchema>,
         partition_filters: Option<Vec<(&str, &str, PartitionFilterValue)>>,
-        schema: ArrowSchema,
     ) -> PyResult<Vec<(String, Option<&'py PyAny>)>> {
         let path_set = match partition_filters {
             Some(filters) => Some(HashSet::<_>::from_iter(
@@ -311,15 +364,41 @@ impl RawDeltaTable {
             .collect()
     }
 
+    fn get_active_partitions(
+        &mut self,
+        partitions_filters: Option<Vec<(&str, &str, PartitionFilterValue)>>,
+    ) -> PyResult<HashSet<(String, Option<String>)>> {
+        let converted_filters = convert_partition_filters(partitions_filters.unwrap_or_default())
+            .map_err(PyDeltaTableError::from_raw)?;
+
+        let add_actions = self
+            ._table
+            .get_state()
+            .get_active_add_actions_by_partitions(&converted_filters)
+            .map_err(PyDeltaTableError::from_raw)?;
+        let active_partitions = add_actions
+            .flat_map(|add| {
+                add.partition_values
+                    .iter()
+                    .map(|i| (i.0.to_owned(), i.1.to_owned()))
+                    .collect::<Vec<_>>()
+            })
+            .collect::<HashSet<_>>();
+        Ok(active_partitions)
+    }
+
     fn create_write_transaction(
         &mut self,
         add_actions: Vec<PyAddAction>,
         mode: &str,
         partition_by: Vec<String>,
-        schema: ArrowSchema,
+        schema: PyArrowType<ArrowSchema>,
+        partitions_filters: Option<Vec<(&str, &str, PartitionFilterValue)>>,
     ) -> PyResult<()> {
         let mode = save_mode_from_str(mode)?;
-        let schema: Schema = (&schema).try_into()?;
+        let schema: Schema = (&schema.0)
+            .try_into()
+            .map_err(PyDeltaTableError::from_arrow)?;
 
         let existing_schema = self
             ._table
@@ -333,8 +412,17 @@ impl RawDeltaTable {
 
         match mode {
             SaveMode::Overwrite => {
-                // Remove all current files
-                for old_add in self._table.get_state().files().iter() {
+                let converted_filters =
+                    convert_partition_filters(partitions_filters.unwrap_or_default())
+                        .map_err(PyDeltaTableError::from_raw)?;
+
+                let add_actions = self
+                    ._table
+                    .get_state()
+                    .get_active_add_actions_by_partitions(&converted_filters)
+                    .map_err(PyDeltaTableError::from_raw)?;
+
+                for old_add in add_actions {
                     let remove_action = Action::remove(action::Remove {
                         path: old_add.path.clone(),
                         deletion_timestamp: Some(current_timestamp()),
@@ -386,11 +474,42 @@ impl RawDeltaTable {
         Ok(())
     }
 
-    pub fn get_py_storage_backend(&self) -> filesystem::DeltaFileSystemHandler {
-        filesystem::DeltaFileSystemHandler {
+    pub fn get_py_storage_backend(&self) -> PyResult<filesystem::DeltaFileSystemHandler> {
+        Ok(filesystem::DeltaFileSystemHandler {
             inner: self._table.object_store(),
-        }
+            rt: Arc::new(rt()?),
+            config: self._config.clone(),
+        })
     }
+
+    pub fn create_checkpoint(&self) -> PyResult<()> {
+        rt()?
+            .block_on(create_checkpoint(&self._table))
+            .map_err(PyDeltaTableError::from_checkpoint)?;
+
+        Ok(())
+    }
+
+    pub fn get_add_actions(&self, flatten: bool) -> PyResult<PyArrowType<RecordBatch>> {
+        Ok(PyArrowType(
+            self._table
+                .get_state()
+                .add_actions_table(flatten)
+                .map_err(PyDeltaTableError::from_raw)?,
+        ))
+    }
+}
+
+fn convert_partition_filters<'a>(
+    partitions_filters: Vec<(&'a str, &'a str, PartitionFilterValue<'a>)>,
+) -> Result<Vec<PartitionFilter<&'a str>>, deltalake::DeltaTableError> {
+    partitions_filters
+        .into_iter()
+        .map(|filter| match filter {
+            (key, op, PartitionFilterValue::Single(v)) => PartitionFilter::try_from((key, op, v)),
+            (key, op, PartitionFilterValue::Multiple(v)) => PartitionFilter::try_from((key, op, v)),
+        })
+        .collect()
 }
 
 fn json_value_to_py(value: &serde_json::Value, py: Python) -> PyObject {
@@ -417,7 +536,7 @@ fn json_value_to_py(value: &serde_json::Value, py: Python) -> PyObject {
 /// skipped during a scan.
 fn filestats_to_expression<'py>(
     py: Python<'py>,
-    schema: &ArrowSchema,
+    schema: &PyArrowType<ArrowSchema>,
     partitions_values: &HashMap<String, Option<String>>,
     stats: Option<Stats>,
 ) -> PyResult<Option<&'py PyAny>> {
@@ -427,14 +546,16 @@ fn filestats_to_expression<'py>(
     let mut expressions: Vec<PyResult<&PyAny>> = Vec::new();
 
     let cast_to_type = |column_name: &String, value: PyObject, schema: &ArrowSchema| {
-        let column_type = schema
-            .field_with_name(column_name)
-            .map_err(|_| {
-                PyDeltaTableError::new_err(format!("Column not found in schema: {}", column_name))
-            })?
-            .data_type()
-            .clone()
-            .into_py(py);
+        let column_type = PyArrowType(
+            schema
+                .field_with_name(column_name)
+                .map_err(|_| {
+                    PyDeltaTableError::new_err(format!("Column not found in schema: {column_name}"))
+                })?
+                .data_type()
+                .clone(),
+        )
+        .into_py(py);
         pa.call_method1("scalar", (value,))?
             .call_method1("cast", (column_type,))
     };
@@ -442,7 +563,7 @@ fn filestats_to_expression<'py>(
     for (column, value) in partitions_values.iter() {
         if let Some(value) = value {
             // value is a string, but needs to be parsed into appropriate type
-            let converted_value = cast_to_type(column, value.into_py(py), schema)?;
+            let converted_value = cast_to_type(column, value.into_py(py), &schema.0)?;
             expressions.push(
                 field
                     .call1((column,))?
@@ -458,7 +579,7 @@ fn filestats_to_expression<'py>(
             // Blocked on https://issues.apache.org/jira/browse/ARROW-11259
             _ => None,
         }) {
-            let maybe_minimum = cast_to_type(&col_name, minimum, schema);
+            let maybe_minimum = cast_to_type(&col_name, minimum, &schema.0);
             if let Ok(minimum) = maybe_minimum {
                 expressions.push(field.call1((col_name,))?.call_method1("__ge__", (minimum,)));
             }
@@ -468,7 +589,7 @@ fn filestats_to_expression<'py>(
             ColumnValueStat::Value(val) => Some((k.clone(), json_value_to_py(val, py))),
             _ => None,
         }) {
-            let maybe_maximum = cast_to_type(&col_name, maximum, schema);
+            let maybe_maximum = cast_to_type(&col_name, maximum, &schema.0);
             if let Ok(maximum) = maybe_maximum {
                 expressions.push(field.call1((col_name,))?.call_method1("__le__", (maximum,)));
             }
@@ -551,15 +672,17 @@ impl From<&PyAddAction> for action::Add {
 #[allow(clippy::too_many_arguments)]
 fn write_new_deltalake(
     table_uri: String,
-    schema: ArrowSchema,
+    schema: PyArrowType<ArrowSchema>,
     add_actions: Vec<PyAddAction>,
     _mode: &str,
     partition_by: Vec<String>,
     name: Option<String>,
     description: Option<String>,
     configuration: Option<HashMap<String, Option<String>>>,
+    storage_options: Option<HashMap<String, String>>,
 ) -> PyResult<()> {
     let mut table = DeltaTableBuilder::from_uri(table_uri)
+        .with_storage_options(storage_options.unwrap_or_default())
         .build()
         .map_err(PyDeltaTableError::from_raw)?;
 
@@ -567,7 +690,9 @@ fn write_new_deltalake(
         name,
         description,
         None, // Format
-        (&schema).try_into()?,
+        (&schema.0)
+            .try_into()
+            .map_err(PyDeltaTableError::from_arrow)?,
         partition_by,
         configuration.unwrap_or_default(),
     );
@@ -610,10 +735,10 @@ impl PyDeltaDataChecker {
         }
     }
 
-    fn check_batch(&self, batch: RecordBatch) -> PyResult<()> {
+    fn check_batch(&self, batch: PyArrowType<RecordBatch>) -> PyResult<()> {
         self.rt.block_on(async {
             self.inner
-                .check_batch(&batch)
+                .check_batch(&batch.0)
                 .await
                 .map_err(PyDeltaTableError::from_raw)
         })
