@@ -2,9 +2,13 @@ import os
 from datetime import datetime
 from pathlib import Path
 from threading import Barrier, Thread
+from unittest.mock import Mock
 
 from packaging import version
 
+from deltalake.table import ProtocolVersions
+from deltalake.writer import DeltaTableProtocolError
+
 try:
     import pandas as pd
 except ModuleNotFoundError:
@@ -156,6 +160,20 @@ def test_read_table_with_column_subset():
     )
 
 
+def test_read_table_as_category():
+    table_path = "../rust/tests/data/delta-0.8.0-partitioned"
+    dt = DeltaTable(table_path)
+
+    assert dt.schema().to_pyarrow().field("value").type == pa.string()
+
+    read_options = ds.ParquetReadOptions(dictionary_columns={"value"})
+
+    data = dt.to_pyarrow_dataset(parquet_read_options=read_options).to_table()
+
+    assert data.schema.field("value").type == pa.dictionary(pa.int32(), pa.string())
+    assert data.schema.field("day").type == pa.string()
+
+
 def test_read_table_with_filter():
     table_path = "../rust/tests/data/delta-0.8.0-partitioned"
     dt = DeltaTable(table_path)
@@ -219,6 +237,15 @@ def test_read_partitioned_table_protocol():
     assert protocol.min_writer_version == 2
 
 
+def test_read_table_with_cdc():
+    table_path = "../rust/tests/data/simple_table_with_cdc"
+    dt = DeltaTable(table_path)
+    assert dt.to_pyarrow_table().to_pydict() == {
+        "id": [0],
+        "name": ["Mino"],
+    }
+
+
 def test_history_partitioned_table_metadata():
     table_path = "../rust/tests/data/delta-0.8.0-partitioned"
     dt = DeltaTable(table_path)
@@ -242,44 +269,104 @@ def test_history_partitioned_table_metadata():
     }
 
 
+@pytest.mark.parametrize("flatten", [True, False])
+def test_add_actions_table(flatten: bool):
+    table_path = "../rust/tests/data/delta-0.8.0-partitioned"
+    dt = DeltaTable(table_path)
+    actions_df = dt.get_add_actions(flatten)
+    # RecordBatch doesn't have a sort_by method yet
+    actions_df = pa.Table.from_batches([actions_df]).sort_by("path").to_batches()[0]
+
+    assert actions_df.num_rows == 6
+    assert actions_df["path"] == pa.array(
+        [
+            "year=2020/month=1/day=1/part-00000-8eafa330-3be9-4a39-ad78-fd13c2027c7e.c000.snappy.parquet",
+            "year=2020/month=2/day=3/part-00000-94d16827-f2fd-42cd-a060-f67ccc63ced9.c000.snappy.parquet",
+            "year=2020/month=2/day=5/part-00000-89cdd4c8-2af7-4add-8ea3-3990b2f027b5.c000.snappy.parquet",
+            "year=2021/month=12/day=20/part-00000-9275fdf4-3961-4184-baa0-1c8a2bb98104.c000.snappy.parquet",
+            "year=2021/month=12/day=4/part-00000-6dc763c0-3e8b-4d52-b19e-1f92af3fbb25.c000.snappy.parquet",
+            "year=2021/month=4/day=5/part-00000-c5856301-3439-4032-a6fc-22b7bc92bebb.c000.snappy.parquet",
+        ]
+    )
+    assert actions_df["size_bytes"] == pa.array([414, 414, 414, 407, 414, 414])
+    assert actions_df["data_change"] == pa.array([True] * 6)
+    assert actions_df["modification_time"] == pa.array(
+        [1615555646000] * 6, type=pa.timestamp("ms")
+    )
+
+    if flatten:
+        partition_year = actions_df["partition.year"]
+        partition_month = actions_df["partition.month"]
+        partition_day = actions_df["partition.day"]
+    else:
+        partition_year = actions_df["partition_values"].field("year")
+        partition_month = actions_df["partition_values"].field("month")
+        partition_day = actions_df["partition_values"].field("day")
+
+    assert partition_year == pa.array(["2020"] * 3 + ["2021"] * 3)
+    assert partition_month == pa.array(["1", "2", "2", "12", "12", "4"])
+    assert partition_day == pa.array(["1", "3", "5", "20", "4", "5"])
+
+
+def assert_correct_files(dt: DeltaTable, partition_filters, expected_paths):
+    assert dt.files(partition_filters) == expected_paths
+    absolute_paths = [os.path.join(dt.table_uri, path) for path in expected_paths]
+    assert dt.file_uris(partition_filters) == absolute_paths
+
+
 def test_get_files_partitioned_table():
     table_path = "../rust/tests/data/delta-0.8.0-partitioned"
     dt = DeltaTable(table_path)
     table_path = (
         Path.cwd().parent / "rust/tests/data/delta-0.8.0-partitioned"
     ).as_posix()
+
     partition_filters = [("day", "=", "3")]
-    assert dt.files_by_partitions(partition_filters=partition_filters) == [
-        f"year=2020/month=2/day=3/part-00000-94d16827-f2fd-42cd-a060-f67ccc63ced9.c000.snappy.parquet"
+    paths = [
+        "year=2020/month=2/day=3/part-00000-94d16827-f2fd-42cd-a060-f67ccc63ced9.c000.snappy.parquet"
     ]
+    assert_correct_files(dt, partition_filters, paths)
+
+    # Also accepts integers
+    partition_filters = [("day", "=", 3)]
+    assert_correct_files(dt, partition_filters, paths)
+
     partition_filters = [("day", "!=", "3")]
-    assert dt.files_by_partitions(partition_filters=partition_filters) == [
-        f"year=2020/month=1/day=1/part-00000-8eafa330-3be9-4a39-ad78-fd13c2027c7e.c000.snappy.parquet",
-        f"year=2020/month=2/day=5/part-00000-89cdd4c8-2af7-4add-8ea3-3990b2f027b5.c000.snappy.parquet",
-        f"year=2021/month=12/day=20/part-00000-9275fdf4-3961-4184-baa0-1c8a2bb98104.c000.snappy.parquet",
-        f"year=2021/month=12/day=4/part-00000-6dc763c0-3e8b-4d52-b19e-1f92af3fbb25.c000.snappy.parquet",
-        f"year=2021/month=4/day=5/part-00000-c5856301-3439-4032-a6fc-22b7bc92bebb.c000.snappy.parquet",
+    paths = [
+        "year=2020/month=1/day=1/part-00000-8eafa330-3be9-4a39-ad78-fd13c2027c7e.c000.snappy.parquet",
+        "year=2020/month=2/day=5/part-00000-89cdd4c8-2af7-4add-8ea3-3990b2f027b5.c000.snappy.parquet",
+        "year=2021/month=12/day=20/part-00000-9275fdf4-3961-4184-baa0-1c8a2bb98104.c000.snappy.parquet",
+        "year=2021/month=12/day=4/part-00000-6dc763c0-3e8b-4d52-b19e-1f92af3fbb25.c000.snappy.parquet",
+        "year=2021/month=4/day=5/part-00000-c5856301-3439-4032-a6fc-22b7bc92bebb.c000.snappy.parquet",
     ]
+    assert_correct_files(dt, partition_filters, paths)
+
     partition_filters = [("day", "in", ["3", "20"])]
-    assert dt.files_by_partitions(partition_filters=partition_filters) == [
-        f"year=2020/month=2/day=3/part-00000-94d16827-f2fd-42cd-a060-f67ccc63ced9.c000.snappy.parquet",
-        f"year=2021/month=12/day=20/part-00000-9275fdf4-3961-4184-baa0-1c8a2bb98104.c000.snappy.parquet",
+    paths = [
+        "year=2020/month=2/day=3/part-00000-94d16827-f2fd-42cd-a060-f67ccc63ced9.c000.snappy.parquet",
+        "year=2021/month=12/day=20/part-00000-9275fdf4-3961-4184-baa0-1c8a2bb98104.c000.snappy.parquet",
     ]
+    assert_correct_files(dt, partition_filters, paths)
+
     partition_filters = [("day", "not in", ["3", "20"])]
-    assert dt.files_by_partitions(partition_filters=partition_filters) == [
-        f"year=2020/month=1/day=1/part-00000-8eafa330-3be9-4a39-ad78-fd13c2027c7e.c000.snappy.parquet",
-        f"year=2020/month=2/day=5/part-00000-89cdd4c8-2af7-4add-8ea3-3990b2f027b5.c000.snappy.parquet",
-        f"year=2021/month=12/day=4/part-00000-6dc763c0-3e8b-4d52-b19e-1f92af3fbb25.c000.snappy.parquet",
-        f"year=2021/month=4/day=5/part-00000-c5856301-3439-4032-a6fc-22b7bc92bebb.c000.snappy.parquet",
+    paths = [
+        "year=2020/month=1/day=1/part-00000-8eafa330-3be9-4a39-ad78-fd13c2027c7e.c000.snappy.parquet",
+        "year=2020/month=2/day=5/part-00000-89cdd4c8-2af7-4add-8ea3-3990b2f027b5.c000.snappy.parquet",
+        "year=2021/month=12/day=4/part-00000-6dc763c0-3e8b-4d52-b19e-1f92af3fbb25.c000.snappy.parquet",
+        "year=2021/month=4/day=5/part-00000-c5856301-3439-4032-a6fc-22b7bc92bebb.c000.snappy.parquet",
     ]
+    assert_correct_files(dt, partition_filters, paths)
+
     partition_filters = [("day", "not in", ["3", "20"]), ("year", "=", "2021")]
-    assert dt.files_by_partitions(partition_filters=partition_filters) == [
-        f"year=2021/month=12/day=4/part-00000-6dc763c0-3e8b-4d52-b19e-1f92af3fbb25.c000.snappy.parquet",
-        f"year=2021/month=4/day=5/part-00000-c5856301-3439-4032-a6fc-22b7bc92bebb.c000.snappy.parquet",
+    paths = [
+        "year=2021/month=12/day=4/part-00000-6dc763c0-3e8b-4d52-b19e-1f92af3fbb25.c000.snappy.parquet",
+        "year=2021/month=4/day=5/part-00000-c5856301-3439-4032-a6fc-22b7bc92bebb.c000.snappy.parquet",
     ]
+    assert_correct_files(dt, partition_filters, paths)
+
     partition_filters = [("invalid_operation", "=>", "3")]
     with pytest.raises(Exception) as exception:
-        dt.files_by_partitions(partition_filters=partition_filters)
+        dt.files(partition_filters)
     assert (
         str(exception.value)
         == 'Invalid partition filter found: ("invalid_operation", "=>", "3").'
@@ -287,26 +374,19 @@ def test_get_files_partitioned_table():
 
     partition_filters = [("invalid_operation", "=", ["3", "20"])]
     with pytest.raises(Exception) as exception:
-        dt.files_by_partitions(partition_filters=partition_filters)
+        dt.files(partition_filters)
     assert (
         str(exception.value)
         == 'Invalid partition filter found: ("invalid_operation", "=", ["3", "20"]).'
     )
 
-    partition_filters = [("day", "=", 3)]
-    with pytest.raises(Exception) as exception:
-        dt.files_by_partitions(partition_filters=partition_filters)
-    assert (
-        str(exception.value)
-        == "Only the type String is currently allowed inside the partition filters."
-    )
-
     partition_filters = [("unknown", "=", "3")]
     with pytest.raises(Exception) as exception:
-        dt.files_by_partitions(partition_filters=partition_filters)
+        dt.files(partition_filters)
     assert (
         str(exception.value)
-        == 'Invalid partition filter found: [PartitionFilter { key: "unknown", value: Equal("3") }].'
+        == 'Invalid partition filter found: [PartitionFilter { key: "unknown", value:'
+        ' Equal("3") }].'
     )
 
 
@@ -325,6 +405,18 @@ def test_delta_table_with_filesystem():
     assert dt.to_pandas(filesystem=filesystem).equals(pd.DataFrame({"id": [5, 7, 9]}))
 
 
+def test_writer_fails_on_protocol():
+    table_path = "../rust/tests/data/simple_table"
+    dt = DeltaTable(table_path)
+    dt.protocol = Mock(return_value=ProtocolVersions(2, 1))
+    with pytest.raises(DeltaTableProtocolError):
+        dt.to_pyarrow_dataset()
+    with pytest.raises(DeltaTableProtocolError):
+        dt.to_pyarrow_table()
+    with pytest.raises(DeltaTableProtocolError):
+        dt.to_pandas()
+
+
 def test_import_delta_table_error():
     from deltalake import PyDeltaTableError
 
@@ -377,8 +469,7 @@ class ExcPassThroughThread(Thread):
 @pytest.mark.integration
 @pytest.mark.timeout(timeout=5, method="thread")
 def test_read_multiple_tables_from_s3(s3_localstack):
-    """
-    Should be able to create multiple cloud storage based DeltaTable instances
+    """Should be able to create multiple cloud storage based DeltaTable instances
     without blocking on async rust function calls.
     """
     for path in ["s3://deltars/simple", "s3://deltars/simple"]:
